/*

nla-ginac - Tarjan-blocking NLA solver with autodiff
Copyright (C) 2005  John Pye

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#include <stdexcept>
#include <cmath>
#include <sstream>

#include "colebrook.h"

double
Colebrook::f_ReEd(const double &Re, const double &epsOnD){

	if(Re <= 3000){
		return 64/Re;
	}

	if(!epsOnD){

		// Smooth pipe with Re > 3000, use Petukov

		if(Re > 5E6){
			throw std::runtime_error("EMSOpipe::frictionFactor: Reynolds number out of range");
		}

		return pow(0.790 * log(Re) - 1.64,-2.0);

	}else{

		// Iteration on Colebrook equation for rough pipes...

		double oneOnRootFnew = 0;
		double oneOnRootFold = 1;

		int iter=0;
		while(1){
			oneOnRootFold = oneOnRootFnew;
			oneOnRootFnew = -0.86 * log(epsOnD/3.7 + 2.51/Re*oneOnRootFold);
			iter++;

			if(fabs(oneOnRootFnew-oneOnRootFold)<=COLEBROOK_TOL){
				break;
			}

			if(iter>=MAX_COLEBROOK_ITER){
				std::stringstream s;
				s << "EMSOpipe::frictionFactor: Unable to converge Colebrook equation for pipe roughness with " << MAX_COLEBROOK_ITER << " iterations: ";
				s << " error now = " << fabs(oneOnRootFnew-oneOnRootFold) << " >  required " << COLEBROOK_TOL;
				throw std::runtime_error(s.str());
			}
		}

		return 1.0 / oneOnRootFnew/oneOnRootFnew;
	}
}
