/*

nla-ginac - Tarjan-blocking NLA solver with autodiff
Copyright (C) 2005  John Pye

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#include "symbolicsystem.h"
#include "nnessolver.h"
#include "roulette.h"
#include "tarjansolver.h"
#include "iapws-sat.h"
#include "units.h"
#include "colebrook-sym.h"
#include "viscosity-sym.h"

using namespace std;
using namespace GiNaC;

class ColebrookSystem: public SymbolicSystem{

	virtual const char *getType() const{return "ColebrookSystem";}

	virtual void defineSystem(){

		CONST(p1,60.0 * bar/MPa);
		CONST(m1,0.01);
		CONST(T1,250+273.15);
		VARIABLE(rho1,802);
		VARIABLE(h1,water_h(T1,802));
		VARIABLE(mu1,1e-5);
		VARIABLE(Re,7e7);

        VARIABLE(f,0.02);
        VARIABLE(DP, double(-10.0 * bar/ MPa));

		VARIABLE(p2,50.0 * bar/MPa);

		VARIABLE(m2,0.05);
		VARIABLE(T2, 180 + 273.15);
		VARIABLE(rho2,0.001);
		VARIABLE(h2, 2000);

		CONST(hpipe,10.0 / kilo);
		CONST(Tamb,300);

        CONST(eps,0.046e-3);
		CONST(L,100);
		CONST(D,0.05);

		CONST(A1,Pi * D*D/4);
		CONST(Apipe, Pi*D*L);

        LOWERB(h2,0);
        LOWERB(rho2,0);
        LOWERB(T2,0);

		lst ff; ff =
			// inlet properties
			p1 - water_p(T1,rho1)
			, h1 - water_h(T1,rho1)

			// outlet properties
			, h2 - water_h(T2,rho2)
			, p2 - water_p(T2,rho2)

			// pressure drop
			, mu1 - mu_Trho(T1,rho1)
            , Re - rho1*m1*4/Pi/D/mu1
            , f - f_ReEd(Re,eps/D)
            , DP + 0.5 * f * L / D * pow(m1,2) / pow(A1,2) / rho1
            , (p2 - p1) - DP

			// energy conservation (conv/cond from outside)
			, m1 / 2 * ( h2 - h1) + hpipe * Apipe * (T1 /2  - Tamb)

			// mass conservation
            , m2 - m1

		;
        setExpressions(ff);

	}

	virtual void applySolverOptions(Solver *solver) const{
		//cerr << "COLEBROOK: SET SOLVER OPTIONS" << endl;
		// do bounds checking in NNES:

        //OPTION(ITSCLX,10);
        //OPTION(ITSCLF,9);
		OPTION(MAXIT,50);
        OPTION(STOPCR,2);
        OPTION(JACTYP,0);
        OPTION(FDTOLJ,8e-3);
        OPTION(OUTPUT,0);
        //OPTION(SHOW_BLOCKS,1);
	}

};

int
main(){
    ColebrookSystem P;
	TarjanSolver *S = new TarjanSolver();

    P.accept(S);

	return 0;
}

