
// Unit tests for IAPWS95symbolic class.  Returns 0 if successful.

#include "iapws-low.h"
#include <ginac/ginac.h>
using namespace GiNaC;

#include <sstream>
#include <stdexcept>
using namespace std;

#define VERBOSE
/*

nla-ginac - Tarjan-blocking NLA solver with autodiff
Copyright (C) 2005  John Pye

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#undef assert
#ifdef VERBOSE
#define assert(EXPR) cerr<<"Assertion: " << #EXPR<< ": " << ((EXPR) ? "OK" : "FALSE") << endl;
#else
#define assert(EXPR) if(!(EXPR))throw runtime_error(#EXPR)
#endif

/**
	Get percent different between parameter values
*/
numeric PctDiff(const ex & a, const ex & b)
{
	if(!is_a<numeric>(a)){
		stringstream s;
		s << "PctDiff argument a = " << a << " is not numeric";
		throw runtime_error(s.str());
	}
	if(!is_a<numeric>(b)){
		stringstream s;
		s << "PctDiff argument b = " << b << " is not numeric";
		throw runtime_error(s.str());
	}

	numeric aa=ex_to<numeric>(a);
	numeric bb=ex_to<numeric>(b);

	if (aa==0 && bb==0){
		return 0.0;
	}

	return 100.0 * abs((aa-bb)/bb);
}


void TestIdealGasPart()
{
	// Test that we match Table 6 values.  Note the phi0tautau
	// expected value is negative, as it has to be, although the
	// table gives it as positive.
	numeric eps = 0.0001;
	numeric T = 500;       // K
	numeric rho = 838.025; // kg/m^3
	IAPWS95symbolic w;
	assert(PctDiff(phi0(          rho/w.rhoc, w.Tc/T),  0.204797734e+1) < eps);
	assert(PctDiff(phi0delta(     rho/w.rhoc, w.Tc/T),  0.384236747e+0) < eps);
	assert(PctDiff(phi0deltadelta(rho/w.rhoc, w.Tc/T), -0.147637878e+0) < eps);
	assert(PctDiff(phi0tau(       rho/w.rhoc, w.Tc/T),  0.904611106e+1) < eps);
	assert(PctDiff(phi0tautau(    rho/w.rhoc, w.Tc/T), -0.193249185e+1) < eps);
	assert(PctDiff(phi0deltatau(  rho/w.rhoc, w.Tc/T),  0.000000000e+0) < eps);
}

void TestRealGasPart()
{
	// Test that we match Table 6 values
	numeric eps = 0.0001;
	numeric T = 500;       // K
	numeric rho = 838.025; // kg/m^3
	IAPWS95symbolic w;
	assert(PctDiff(phir(          rho/w.rhoc, w.Tc/T), -0.342693206e+1) < eps);
	assert(PctDiff(phirdelta(     rho/w.rhoc, w.Tc/T), -0.364366650e+0) < eps);
	assert(PctDiff(phirdeltadelta(rho/w.rhoc, w.Tc/T),  0.856063701e+0) < eps);
	assert(PctDiff(phirtau(       rho/w.rhoc, w.Tc/T), -0.581403435e+1) < eps);
	assert(PctDiff(phirtautau(    rho/w.rhoc, w.Tc/T), -0.223440737e+1) < eps);
	assert(PctDiff(phirdeltatau(  rho/w.rhoc, w.Tc/T), -0.112176915e+1) < eps);
}

void TestSinglePhaseRegion()
{
	// Values from Table 7
	numeric values[11][6] = {
		//Temp   Density       Pressure     Isochoric Sp. Ht  Spd of Sound       Entropy
		// K     kg/m^3           MPa          kJ/(kg*K)          m/s           kJ/(kg*K)
		// T       rho             p              cv               w                s
		{300.0, 0.9965560e+3, 0.99242e-1    , 0.413018111e+1, 0.150151914e+4, 0.393062642e+0},
		{300.0, 0.1005308e+4, 0.200022514e+2, 0.406798347e+1, 0.153492501e+4, 0.387405401e+0},
		{300.0, 0.1188202e+4, 0.700004704e+3, 0.346135580e+1, 0.244357992e+4, 0.132609616e+0},
		{500.0, 0.4350000e+0, 0.999679423e-1, 0.150817541e+1, 0.548314253e+3, 0.794488271e+1},
		{500.0, 0.4532000e+1, 0.999938125e+0, 0.166991025e+1, 0.535739001e+3, 0.682502725e+1},
		{500.0, 0.8380250e+3, 0.100003858e+2, 0.322106219e+1, 0.127128441e+4, 0.256690918e+1},
		{500.0, 0.1084564e+4, 0.700000405e+3, 0.307437693e+1, 0.241200877e+4, 0.203237509e+1},
		{647.0, 0.3580000e+3, 0.220384756e+2, 0.618315728e+1, 0.252145078e+3, 0.432092307e+1},
		{900.0, 0.2410000e+0, 0.100062559e+0, 0.175890657e+1, 0.724027147e+3, 0.916653194e+1},
		{900.0, 0.5261500e+2, 0.200000690e+2, 0.193510526e+1, 0.698445674e+3, 0.659070225e+1},
		{900.0, 0.8707690e+3, 0.700000006e+3, 0.266422350e+1, 0.201933608e+4, 0.417223802e+1}
	};
	for (int i = 0; i < 11; i++)
	{
		IAPWS95symbolic w;
		numeric T, rho, p_e, cv_e, w_e, s_e, eps;
		T    = values[i][0];
		rho  = values[i][1];
		p_e  = values[i][2];
		cv_e = values[i][3];
		w_e  = values[i][4];
		s_e  = values[i][5];

		if (T == 647.0)
			eps = 0.2; // Precision at the critical point is much less
		else
			eps = 0.0002;

		assert(PctDiff(w.p(T, rho), p_e) < eps);
		assert(PctDiff(w.cv(T, rho), cv_e) < eps);
		assert(PctDiff(w.w(T, rho), w_e) < eps);
		assert(PctDiff(w.s(T, rho), s_e) < eps);
	}

	// Tests for u and h came from NIST webpage at
	// http://webbook.nist.gov/chemistry/fluid/
	numeric values1[6][5] = {
		// T    rho         u       h        cp
		{350, 973.70,     321.75, 321.79,  4.1946},  // Liquid
		{400, 937.49,     532.69, 532.95,  4.2555},  // Liquid
		{440, 900.65,     704.50, 705.31,  4.3571},  // Liquid
		{350,   0.26029, 2477.6, 2637.7,   2.0033},  // Vapor
		{400,   1.3694,  2536.2, 2715.7,   2.2183},  // Vapor
		{500,   4.5320,  2670.6, 2891.2,   2.2795}   // Vapor
	};
	for (int i = 0; i < 6; i++)
	{
		IAPWS95symbolic w;
		numeric T, rho, u_e, h_e, cp_e;
		numeric eps = 0.0025;
		T    = values1[i][0];
		rho  = values1[i][1];
		u_e  = values1[i][2];
		h_e  = values1[i][3];
		cp_e = values1[i][4];
		assert(PctDiff(w.u(T, rho), u_e) < eps);
		assert(PctDiff(w.h(T, rho), h_e) < eps);
		assert(PctDiff(w.cp(T, rho), cp_e) < eps);
	}
}

int main(){
	try{
		TestIdealGasPart();
		TestRealGasPart();
		TestSinglePhaseRegion();
	}catch(exception &e){
		cerr << "ERROR: " << e.what() << endl;
		exit(1);
	}
	exit(0);
}
