/*

nla-ginac - Tarjan-blocking NLA solver with autodiff
Copyright (C) 2005  John Pye

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#include <iostream>
#include <stdexcept>

#include "iapws-low.h"
#include "iapws95.h"
#include "saturation.h"

using namespace GiNaC;
using namespace std;

#define REGISTER_EVAL_DIFF(FN) REGISTER_FUNCTION(FN,eval_func(FN ## _eval).derivative_func(FN ## _diff))

namespace IAPWS95_utils
{

	/// SATURATION CURVES

	///---------------
	static ex deltaf_eval(const ex & tau){
		ex tau1=tau.evalf();
		if(is_a<numeric>(tau1)){
			double t = ex_to<numeric>(tau1).to_double();
			return Sat::deltaf(t);
		}else{
			return deltaf(tau).hold();
		}
	}

	static ex deltaf_diff(const ex & tau, unsigned diff_param){
		return deltaftau(tau);
	}

	REGISTER_EVAL_DIFF(deltaf);

	///--------------

	static ex deltaftau_eval(const ex & tau){
		ex tau1=tau.evalf();
		if(is_a<numeric>(tau1)){
			double t = ex_to<numeric>(tau1).to_double();
			return Sat::deltaftau(t);
		}else{
			return deltaftau(tau).hold();
		}
	}

	REGISTER_FUNCTION(deltaftau,eval_func(deltaftau_eval));

	///---------------
	static ex deltag_eval(const ex & tau){
		ex tau1=tau.evalf();
		if(is_a<numeric>(tau1)){
			double t = ex_to<numeric>(tau1).to_double();
			return Sat::deltag(t);
		}else{
			return deltag(tau).hold();
		}
	}

	static ex deltag_diff(const ex & tau, unsigned diff_param){
		return deltagtau(tau);
	}

	REGISTER_EVAL_DIFF(deltag);

	///--------------
	static ex deltagtau_eval(const ex & tau){
		ex tau1=tau.evalf();
		if(is_a<numeric>(tau1)){
			double t = ex_to<numeric>(tau1).to_double();
			return Sat::deltagtau(t);
		}else{
			return deltagtau(tau).hold();
		}
	}

	REGISTER_FUNCTION(deltagtau,eval_func(deltagtau_eval));

	///--------------


	/// PARTIAL DERIVATIVES

	static ex phi0_eval(const ex & delta, const ex & tau)
	{
		ex tau1=tau.evalf();
		ex delta1=delta.evalf();
		if(is_a<numeric>(delta1) && is_a<numeric>(tau1)){
			double d = ex_to<numeric>(delta1).to_double();
			double t = ex_to<numeric>(tau1).to_double();
			return IAPWS95::phi0(d,t);
		}else{
			return phi0(delta,tau).hold();
		}
	}
	static ex phi0_diff(const ex &delta, const ex &tau, unsigned diff_param){
		switch(diff_param){
			case 0:
				return phi0delta(delta,tau);
			case 1:
				return phi0tau(delta,tau);
			default:
				throw runtime_error("phi0_diff: invalid diff_param");
		}
	}

	REGISTER_EVAL_DIFF(phi0);

	///----------
	static ex phi0delta_eval(const ex & delta, const ex & tau)
	{
		ex tau1=tau.evalf();
		ex delta1=delta.evalf();
		if(is_a<numeric>(delta1) && is_a<numeric>(tau1)){
			double d=ex_to<numeric>(delta1).to_double();
			double t=ex_to<numeric>(tau1).to_double();
			return IAPWS95::phi0delta(d,t);
		}else{
			return phi0delta(delta,tau).hold();
		}
	}
	static ex phi0delta_diff(const ex &delta, const ex &tau, unsigned diff_param){
		switch(diff_param){
			case 0:
				return phi0deltadelta(delta,tau);
			case 1:
				return phi0deltatau(delta,tau);
			default:
				throw runtime_error("phi0delta_diff: invalid diff_param");
		}
	}


	REGISTER_EVAL_DIFF(phi0delta);

	///-----------
	static ex phi0tau_eval(const ex & delta, const ex & tau)
	{
		ex tau1=tau.evalf();
		ex delta1=delta.evalf();
		if(is_a<numeric>(delta1) && is_a<numeric>(tau1)){
			double d=ex_to<numeric>(delta1).to_double();
			double t=ex_to<numeric>(tau1).to_double();
			return IAPWS95::phi0tau(d,t);
		}else{
			return phi0tau(delta,tau).hold();
		}
	}
	static ex phi0tau_diff(const ex &delta, const ex &tau, unsigned diff_param){
		switch(diff_param){
			case 0:
				return phi0deltatau(delta,tau);
			case 1:
				return phi0tautau(delta,tau);
			default:
				throw runtime_error("phi0tau_diff: invalid diff_param");
		}
	}

	REGISTER_EVAL_DIFF(phi0tau);

	///------------
	static ex phi0deltadelta_eval(const ex &delta, const ex &tau){
		return -1/delta/delta;
	}
	static ex phi0deltadelta_diff(const ex &delta, const ex &tau, unsigned diff_param){
		switch(diff_param){
			case 0:
				return 2/pow(delta,3);
			case 1:
				return 0;
			default:
				throw runtime_error("phi0deltadelta_diff: invalid diff_param");
		}
	}

	REGISTER_EVAL_DIFF(phi0deltadelta);


	///------------
	static ex phi0deltatau_eval(const ex &delta, const ex &tau){
		return 0.0;
	}
	static ex phi0deltatau_diff(const ex & delta, const ex & tau, unsigned diff_param){
		return 0;
	}

	REGISTER_EVAL_DIFF(phi0deltatau);

	///------------
	static ex phi0tautau_eval(const ex & delta, const ex & tau)
	{
		ex tau1=tau.evalf();
		ex delta1=delta.evalf();
		if(is_a<numeric>(delta1) && is_a<numeric>(tau1)){
			double d=ex_to<numeric>(delta1).to_double();
			double t=ex_to<numeric>(tau1).to_double();
			return IAPWS95::phi0tautau(d,t);
		}else{
			return phi0tautau(delta,tau).hold();
		}
	}
	static ex phi0tautau_diff(const ex & delta, const ex & tau, unsigned diff_param){
		switch(diff_param){
			case 0:
				return 0;
			case 1:
				return phi0tautautau(delta,tau);
			default:
				throw runtime_error("phi0tautau_diff: invalid diff_param");
		}
	}

	REGISTER_EVAL_DIFF(phi0tautau);

	///------------
	static ex phi0tautautau_eval(const ex & delta, const ex & tau){

		/// @TODO, implement this:
		return phi0tautautau(delta,tau).hold();
	}

	REGISTER_FUNCTION(phi0tautautau,eval_func(phi0tautautau_eval));

	/**
		RESIDUAL PARTS... a bit harder...
	*/

	///------------
	static ex phir_eval(const ex & delta, const ex & tau)
	{
		ex tau1=tau.evalf();
		ex delta1=delta.evalf();
		if(is_a<numeric>(delta1) && is_a<numeric>(tau1)){
			double d=ex_to<numeric>(delta1).to_double();
			double t=ex_to<numeric>(tau1).to_double();
			return IAPWS95::phir(d,t);
		}else{
			return phir(delta,tau).hold();
		}
	}
	static ex phir_diff(const ex &delta, const ex &tau, unsigned diff_param){
		switch(diff_param){
			case 0:
				return phirdelta(delta,tau);
			case 1:
				return phirtau(delta,tau);
			default:
				throw runtime_error("phir_diff: invalid diff_param");
		}
	}

	REGISTER_EVAL_DIFF(phir);


	///------------
	static ex phirdelta_eval(const ex & delta, const ex & tau)
	{
		ex tau1=tau.evalf();
		ex delta1=delta.evalf();
		if(is_a<numeric>(delta1) && is_a<numeric>(tau1)){
			double d=ex_to<numeric>(delta1).to_double();
			double t=ex_to<numeric>(tau1).to_double();
			return IAPWS95::phirdelta(d,t);
		}else{
			return phirdelta(delta,tau).hold();
		}
	}
	static ex phirdelta_diff(const ex &delta, const ex &tau, unsigned diff_param){
		switch(diff_param){
			case 0:
				return phirdeltadelta(delta,tau);
			case 1:
				return phirdeltatau(delta,tau);
			default:
				throw runtime_error("phirdelta_diff: invalid diff_param");
		}
	}

	REGISTER_EVAL_DIFF(phirdelta);


	///------------
	static ex phirtau_eval(const ex & delta, const ex & tau)
	{
		ex tau1=tau.evalf();
		ex delta1=delta.evalf();
		if(is_a<numeric>(delta1) && is_a<numeric>(tau1)){
			double d=ex_to<numeric>(delta1).to_double();
			double t=ex_to<numeric>(tau1).to_double();
			return IAPWS95::phirtau(d,t);
		}else{
			return phirtau(delta,tau).hold();
		}
	}
	static ex phirtau_diff(const ex &delta, const ex &tau, unsigned diff_param){
		switch(diff_param){
			case 0:
				return phirdeltatau(delta,tau);
			case 1:
				return phirtautau(delta,tau);
			default:
				throw runtime_error("phirtau_diff: invalid diff_param");
		}
	}

	REGISTER_EVAL_DIFF(phirtau);

	///------------
	static ex phirdeltadelta_eval(const ex & delta, const ex & tau)
	{
		ex tau1=tau.evalf();
		ex delta1=delta.evalf();
		if(is_a<numeric>(delta1) && is_a<numeric>(tau1)){
			double d=ex_to<numeric>(delta1).to_double();
			double t=ex_to<numeric>(tau1).to_double();
			return IAPWS95::phirdeltadelta(d,t);
		}else{
			return phirdeltadelta(delta,tau).hold();
		}
	}
	static ex phirdeltadelta_diff(const ex &delta, const ex &tau, unsigned diff_param){
		switch(diff_param){
			case 0:
				return phirdeltadeltadelta(delta,tau);
			case 1:
				return phirdeltadeltatau(delta,tau);
			default:
				throw runtime_error("phirdeltadelta_diff: invalid diff_param");
		}
	}


	REGISTER_EVAL_DIFF(phirdeltadelta);

	///-------------
	static ex phirdeltatau_eval(const ex & delta, const ex & tau)
	{
		ex tau1=tau.evalf();
		ex delta1=delta.evalf();
		if(is_a<numeric>(delta1) && is_a<numeric>(tau1)){
			double d=ex_to<numeric>(delta1).to_double();
			double t=ex_to<numeric>(tau1).to_double();
			return numeric(IAPWS95::phirdeltatau(d,t));
		}else{
			return phirdeltatau(delta,tau).hold();
		}
	}

	static ex phirdeltatau_diff(const ex &delta, const ex &tau, unsigned diff_param){
		switch(diff_param){
			case 0:
				return phirdeltadeltatau(delta,tau);
			case 1:
				return phirdeltatautau(delta,tau);
			default:
				throw runtime_error("phirdeltatau_diff: invalid diff_param");
		}
	}


	REGISTER_EVAL_DIFF(phirdeltatau);

	///-------------
	static ex phirtautau_eval(const ex & delta, const ex & tau)
	{
		ex tau1=tau.evalf();
		ex delta1=delta.evalf();
		if(is_a<numeric>(delta1) && is_a<numeric>(tau1)){
			double d=ex_to<numeric>(delta1).to_double();
			double t=ex_to<numeric>(tau1).to_double();
			return IAPWS95::phirtautau(d,t);
		}else{
			return phirtautau(delta,tau).hold();
		}
	}

	static ex phirtautau_diff(const ex &delta, const ex &tau, unsigned diff_param){
		switch(diff_param){
			case 0:
				return phirdeltatautau(delta,tau);
			case 1:
				return phirtautautau(delta,tau);
			default:
				throw runtime_error("phirtautau_diff: invalid diff_param");
		}
	}


	REGISTER_EVAL_DIFF(phirtautau);

	///-------------
	/// DUMMY FUNCTIONS, remain to be implemented, if/when nec

	REGISTER_FUNCTION(phirdeltadeltadelta,dummy());
	REGISTER_FUNCTION(phirdeltadeltatau,dummy());
	REGISTER_FUNCTION(phirdeltatautau,dummy());
	REGISTER_FUNCTION(phirtautautau,dummy());

	/// Convert delta/tau to non-dimensional values
	inline void get_delta_tau(const ex & T, const ex & rho, ex & delta, ex & tau)
	{
		delta = rho/rhoc_;
		tau   = Tc_/T;
	}

} // end namespace IAPWS95_untils

using namespace IAPWS95_utils;

// Nondimensionalising, IAPWS-95

IAPWS95symbolic::IAPWS95symbolic()
    :
    R(R_),
    Tc(Tc_),
    rhoc(rhoc_)
{
    // Intentionally empty
}

// Saturation curves

ex IAPWS95symbolic::rhof(const ex & T)
{
    //std::cerr << "T = " << T << std::endl;
    ex tau = 1 - T / Sat::T_c;
    //std::cerr<< "TAU = " << tau << std::endl;
    ex deltaf = IAPWS95_utils::deltaf(tau);
    //std::cerr<< "DELTAF = " <<deltaf << std::endl;
    return deltaf * Sat::rho_c;
}

ex IAPWS95symbolic::rhog(const ex &T){
    ex tau = 1 - T / Sat::T_c;
    //std::cerr<< "TAU = " << tau << std::endl;
    ex deltag = IAPWS95_utils::deltag(tau);
    //std::cerr<< "DELTAG = " <<deltag << std::endl;
    //std::cerr << "RHOC = " << Sat::rho_c << std::endl;
    //std::cerr << "RHOG = " << deltag * Sat::rho_c << std::endl;
    return deltag * Sat::rho_c;
}

// Property correlations

ex IAPWS95symbolic::p(const ex & T, const ex & rho)
{
    ex delta;
    ex tau;
    get_delta_tau(T, rho, delta, tau);
    return rho*R*T*(1 + delta*phirdelta(delta, tau))/1000;
}

ex IAPWS95symbolic::u(const ex & T, const ex & rho)
{
    ex delta;
    ex tau;
    get_delta_tau(T, rho, delta, tau);
    return R*T*tau*(phi0tau(delta, tau) +
            phirtau(delta, tau));
}

ex IAPWS95symbolic::h(const ex & T, const ex & rho)
{
    ex delta;
    ex tau;
    get_delta_tau(T, rho, delta, tau);
    return R*T*(1 + tau*(phi0tau(delta, tau) +
            phirtau(delta, tau)) + delta*phirdelta(delta, tau));
}

ex IAPWS95symbolic::s(const ex & T, const ex & rho)
{
    ex delta;
    ex tau;
    get_delta_tau(T, rho, delta, tau);
    return R*(tau*(phi0tau(delta, tau) +
            phirtau(delta, tau)) -
            phi0(delta, tau) - phir(delta, tau));
}

ex IAPWS95symbolic::cv(const ex & T, const ex & rho)
{
    ex delta;
    ex tau;
    get_delta_tau(T, rho, delta, tau);
    return -R*tau*tau*(phi0tautau(delta, tau) +
            phirtautau(delta, tau));
}

ex IAPWS95symbolic::cp(const ex & T, const ex & rho)
{
    ex delta;
    ex tau;
    get_delta_tau(T, rho, delta, tau);
    ex a1 = 1 + delta*phirdelta(delta, tau) -
        delta*tau*phirdeltatau(delta, tau);
    ex b1 = 1 + 2*delta*phirdelta(delta, tau) +
        delta*delta*phirdeltadelta(delta, tau);
    return R*(-tau*tau*(phi0tautau(delta, tau) +
            phirtautau(delta, tau)) + a1*a1/b1);
}

ex IAPWS95symbolic::w(const ex & T, const ex & rho)
{
    ex delta;
    ex tau;
    get_delta_tau(T, rho, delta, tau);
    ex a1 = 1 + delta*phirdelta(delta, tau) -
        delta*tau*phirdeltatau(delta, tau);
    ex b1 = tau*tau*(phi0tautau(delta, tau) +
        phirtautau(delta, tau));
    ex w = 1 + 2*delta*phirdelta(delta, tau) +
        delta*delta*phirdeltadelta(delta, tau) - a1*a1/b1;
    return sqrt(R*T*w*1000);
}

