/*

nla-ginac - Tarjan-blocking NLA solver with autodiff
Copyright (C) 2005  John Pye

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
/**
	Fast implementations of low-level routines from IAPWS-95
	Allows IAPWS-95 to be implemented in GiNaC without intolerable loss of speed
*/

#ifndef IAPWS_LOW_H
#define IAPWS_LOW_H

#include <ginac/ginac.h>

namespace IAPWS95_utils
{

	/**
		Saturation curves, dimensionless
	*/

	DECLARE_FUNCTION_1P(deltaf);
	DECLARE_FUNCTION_1P(deltag);
	DECLARE_FUNCTION_1P(deltaftau);
	DECLARE_FUNCTION_1P(deltagtau);

	/**
		These are the low-level functions from IAPWS-95
	*/

	DECLARE_FUNCTION_2P(phi0);
	DECLARE_FUNCTION_2P(phi0delta);
	DECLARE_FUNCTION_2P(phi0tau);
	DECLARE_FUNCTION_2P(phi0deltadelta);
	DECLARE_FUNCTION_2P(phi0deltatau);
	DECLARE_FUNCTION_2P(phi0tautau);

	/// third derivative, not implemented, will all 'hold':
	DECLARE_FUNCTION_2P(phi0tautautau);

	DECLARE_FUNCTION_2P(phir);
	DECLARE_FUNCTION_2P(phirdelta);
	DECLARE_FUNCTION_2P(phirtau);
	DECLARE_FUNCTION_2P(phirdeltadelta);
	DECLARE_FUNCTION_2P(phirtautau);
	DECLARE_FUNCTION_2P(phirdeltatau);

	/// third derivatives, not implemented, will all 'hold':
	DECLARE_FUNCTION_2P(phirtautautau);
	DECLARE_FUNCTION_2P(phirdeltatautau);
	DECLARE_FUNCTION_2P(phirdeltadeltatau);
	DECLARE_FUNCTION_2P(phirdeltadeltadelta);

	/**
		Here is the high-level stuff from IAPWS-95, implemented at the GiNaC level, so that
		derivatives like dp/drho pass down to the lower level courtesy of GiNaC. There are
		not loop constructs in the following, so it should be quite efficient still.
	*/

	// Universal gas constant in kJ/(kg*K)
	const GiNaC::numeric R_    = 0.46151805;

	// Critical temperature of water in K
	const GiNaC::numeric Tc_   = 647.096;

	// Critical density of water in kg/m^3
	const GiNaC::numeric rhoc_ = 322;

	inline void get_delta_tau(const GiNaC::ex & T, const GiNaC::ex & rho, GiNaC::ex & delta, GiNaC::ex & tau);

} // end namespace IAPWS95_utils

class IAPWS95symbolic
{
	public:
		IAPWS95symbolic();
		~IAPWS95symbolic() {}

		static GiNaC::ex rhof(const GiNaC::ex &T);
		static GiNaC::ex rhog(const GiNaC::ex &T);

		GiNaC::ex  p(const GiNaC::ex & T, const GiNaC::ex & rho);
		GiNaC::ex  u(const GiNaC::ex & T, const GiNaC::ex & rho);
		GiNaC::ex  h(const GiNaC::ex & T, const GiNaC::ex & rho);
		GiNaC::ex  s(const GiNaC::ex & T, const GiNaC::ex & rho);
		GiNaC::ex cp(const GiNaC::ex & T, const GiNaC::ex & rho);
		GiNaC::ex cv(const GiNaC::ex & T, const GiNaC::ex & rho);
		GiNaC::ex  w(const GiNaC::ex & T, const GiNaC::ex & rho);

		// Constants
		const GiNaC::numeric & R;     ///< Universal gas constant, kJ/(kg*K)
		const GiNaC::numeric & Tc;    ///< Water critical temperature, K
		const GiNaC::numeric & rhoc;  ///< Water critical density, kg/m^3
};

#endif // IAPWS_LOW_H
