/*

nla-ginac - Tarjan-blocking NLA solver with autodiff
Copyright (C) 2005  John Pye

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
/**
	This file implements clever symbolic steam tables that
	behave correctly in saturated regions, but only once
	the numerical values of their arguments are known.
*/

#include "iapws-sat.h"

#include "iapws-low.h"
#include "saturation.h"
#include "steamexpr.h"

#include <ginac/ginac.h>
#include <iostream>

using namespace GiNaC;
using namespace std;


/// SATURATION STEAM TABLES: ENTHALPY

static ex water_h_eval(const ex &T, const ex &rho){
	ex T1 = T.evalf();
	ex rho1 = rho.evalf();
	if(is_a<numeric>(T1) && is_a<numeric>(rho1)){
		// return the correlation
		if(T1 < Sat::T_c){

			ex rhof1 = IAPWS95symbolic::rhof(T1);
			ex rhog1 = IAPWS95symbolic::rhog(T1);

			if(rho > rhog1 && rho < rhof1){
				// saturated... interpolate
                //cerr << "SATURATED h(rhof = " << rhof1 << ", rhog = " << rhog1 << ", rho = " << rho <<") ";
				return SteamExpr<'h'>::Instance()->get_sat(T,rho);
			}else{
                //cerr << "SUB/SUPER h ";
            }
		}else{
            //cerr << "SUPERCRIT h ";
        }

		// not saturated
		return SteamExpr<'h'>::Instance()->get(T,rho);

	}else{
		return water_h(T,rho).hold();
	}
}

static ex water_h_diff(const ex &T, const ex &rho, unsigned diff_param){
	switch(diff_param){
		case 0:
			return water_h_dT(T,rho);
		case 1:
			return water_h_drho(T,rho);
		default:
			throw runtime_error("water_h_diff: invalid diff_param");
	}
}

REGISTER_FUNCTION(water_h, eval_func(water_h_eval).derivative_func(water_h_diff));

static ex water_h_dT_eval(const ex &T, const ex &rho){
	ex T1 = T.evalf();
	ex rho1 = rho.evalf();
	if(is_a<numeric>(T1) && is_a<numeric>(rho1)){
		// return the correlation
		if(T1 < Sat::T_c){
			ex rhof1 = IAPWS95symbolic::rhof(T1);
			ex rhog1 = IAPWS95symbolic::rhog(T1);

			if(rho > rhog1 && rho < rhof1){
				// saturated
				return SteamExpr<'h'>::Instance()->get_dT_sat(T,rho);
			}
		}

		// not saturated
		return SteamExpr<'h'>::Instance()->get_dT(T,rho);

	}else{
		return water_h_dT(T,rho).hold();
	}
}

REGISTER_FUNCTION(water_h_dT, eval_func(water_h_dT_eval));

static ex water_h_drho_eval(const ex &T, const ex &rho){
	ex T1 = T.evalf();
	ex rho1 = rho.evalf();
	if(is_a<numeric>(T1) && is_a<numeric>(rho1)){
		// return the correlation
		if(T1 < Sat::T_c){
			ex rhof1 = IAPWS95symbolic::rhof(T1);
			ex rhog1 = IAPWS95symbolic::rhog(T1);

			if(rho > rhog1 && rho < rhof1){
				// saturated
				//cerr << "SAT ";
				return SteamExpr<'h'>::Instance()->get_drho_sat(T,rho);
			}else if(rho <= rhog1){
                //cerr << "SUP ";
			}else{
                //cerr << "SUB ";
            }
		}else{
            //cerr << "SUPC";
        }

		return SteamExpr<'h'>::Instance()->get_drho(T,rho);

	}else{
		return water_h_drho(T,rho).hold();
	}
}
REGISTER_FUNCTION(water_h_drho, eval_func(water_h_drho_eval));




/// SATURATION STEAM TABLES: PRESSURE

static ex water_p_eval(const ex &T, const ex &rho){
	ex T1 = T.evalf();
	ex rho1 = rho.evalf();
	if(is_a<numeric>(T.evalf()) && is_a<numeric>(rho.evalf())){
		// return the correlation
		if(T1 < Sat::T_c){
			ex rhof1 = IAPWS95symbolic::rhof(T1);
			ex rhog1 = IAPWS95symbolic::rhog(T1);

			if(rho > rhog1 && rho < rhof1){
				// saturated... interpolate
				return SteamExpr<'p'>::Instance()->get_sat(T,rho);
			}

		}
		// not saturated
			return SteamExpr<'p'>::Instance()->get(T,rho);
	}else{
		return water_p(T,rho).hold();
	}
}

static ex water_p_diff(const ex &T, const ex &rho, unsigned diff_param){
	switch(diff_param){
		case 0:
			return water_p_dT(T,rho);
		case 1:
			return water_p_drho(T,rho);
		default:
			throw runtime_error("water_p_diff: invalid diff_param");
	}
}

REGISTER_FUNCTION(water_p, eval_func(water_p_eval).derivative_func(water_p_diff));

static ex water_p_dT_eval(const ex &T, const ex &rho){
	ex T1 = T.evalf();
	ex rho1 = rho.evalf();
	if(is_a<numeric>(T1) && is_a<numeric>(rho1)){
		// return the correlation
		if(T1 < Sat::T_c){
			ex rhof1 = IAPWS95symbolic::rhof(T1);
			ex rhog1 = IAPWS95symbolic::rhog(T1);

			if(rho > rhog1 && rho < rhof1){
				// saturated
				return SteamExpr<'p'>::Instance()->get_dT_sat(T,rho);
			}
		}

		// not saturated
		return SteamExpr<'p'>::Instance()->get_dT(T,rho);

	}else{
		return water_p_dT(T,rho).hold();
	}
}

REGISTER_FUNCTION(water_p_dT, eval_func(water_p_dT_eval));

static ex water_p_drho_eval(const ex &T, const ex &rho){
	ex T1 = T.evalf();
	ex rho1 = rho.evalf();
	if(is_a<numeric>(T1) && is_a<numeric>(rho1)){
		// return the correlation
		if(T1 < Sat::T_c){
			ex rhof1 = IAPWS95symbolic::rhof(T1);
			ex rhog1 = IAPWS95symbolic::rhog(T1);

			if(rho > rhog1 && rho < rhof1){
				// saturated
				return SteamExpr<'p'>::Instance()->get_drho_sat(T,rho);
			}
		}
		// not saturated
		return SteamExpr<'p'>::Instance()->get_drho(T,rho);
	}else{
		return water_p_drho(T,rho).hold();
	}
}
REGISTER_FUNCTION(water_p_drho, eval_func(water_p_drho_eval));






/// SATURATION STEAM TABLES: PRESSURE

static ex water_s_eval(const ex &T, const ex &rho){
	ex T1 = T.evalf();
	ex rho1 = rho.evalf();
	if(is_a<numeric>(T1) && is_a<numeric>(rho1)){
		// return the correlation
		if(T1 < Sat::T_c){
			ex rhof1 = IAPWS95symbolic::rhof(T1);
			ex rhog1 = IAPWS95symbolic::rhog(T1);
			if(rho > rhog1 && rho < rhof1){
				// saturated... interpolate
				//cerr << "SATURATED "<< endl;
				return SteamExpr<'s'>::Instance()->get_sat(T,rho);
			}
		}
		// not saturated
		return SteamExpr<'s'>::Instance()->get(T,rho);
	}else{
		return water_s(T,rho).hold();
	}
}

static ex water_s_diff(const ex &T, const ex &rho, unsigned diff_param){
	switch(diff_param){
		case 0:
			return water_s_dT(T,rho);
		case 1:
			return water_s_drho(T,rho);
		default:
			throw runtime_error("water_s_diff: invalid diff_param");
	}
}

REGISTER_FUNCTION(water_s, eval_func(water_s_eval).derivative_func(water_s_diff));

static ex water_s_dT_eval(const ex &T, const ex &rho){
	ex T1 = T.evalf();
	ex rho1 = rho.evalf();
	if(is_a<numeric>(T1) && is_a<numeric>(rho1)){
		// return the correlation
		if(T1 < Sat::T_c){
			ex rhof1 = IAPWS95symbolic::rhof(T1);
			ex rhog1 = IAPWS95symbolic::rhog(T1);

			if(rho > rhog1 && rho < rhof1){
				// saturated
				return SteamExpr<'s'>::Instance()->get_dT_sat(T,rho);
			}
		}
		// not saturated
		return SteamExpr<'s'>::Instance()->get_dT(T,rho);
	}else{
		return water_s_dT(T,rho).hold();
	}
}

REGISTER_FUNCTION(water_s_dT, eval_func(water_s_dT_eval));

static ex water_s_drho_eval(const ex &T, const ex &rho){
	ex T1 = T.evalf();
	ex rho1 = rho.evalf();
	if(is_a<numeric>(T1) && is_a<numeric>(rho1)){
		// return the correlation
		if(T1 < Sat::T_c){
			ex rhof1 = IAPWS95symbolic::rhof(T1);
			ex rhog1 = IAPWS95symbolic::rhog(T1);

			if(rho > rhog1 && rho < rhof1){
				// saturated
				return SteamExpr<'s'>::Instance()->get_drho_sat(T,rho);
			}
			//cerr << "SUP/SUPER ";
		}else{
			//cerr << "SUPERCRIT ";
		}

		// not saturated
		return SteamExpr<'s'>::Instance()->get_drho(T,rho);
	}else{
		return water_s_drho(T,rho).hold();
	}
}
REGISTER_FUNCTION(water_s_drho, eval_func(water_s_drho_eval));
