#include "matbool.h"

#include <iostream>
#include <stdexcept>

using namespace std;

/*
c
c description of parameters.
c input variables  .... N,ICN,LICN,IP,LENR.
c output variables  ior,ib,num.
C
c N   order of the matrix.
c ICN array containing the column indices of the non-zeros.  those
c     belonging to a single row must be contiguous but the ordering
c     of column indices within each row is unimportant and wasted
c     space between rows is permitted.
c LICN  length of array icn.
c IP  ip(i), i=1,2,...n, is the position in array icn of the first
c     column index of a non-zero in row i.
c LENR  lenr(i) is the number of non-zeros in row i, i=1,2,...n.
c IOR  ior(i) gives the position in the original ordering of the row
c     or column which is in position i in the permuted form, i=1,2,..n.
c IB  ib(i) is the row number in the permuted matrix of the beginning
c     of block i, i=1,2,...num.
c NUM number of blocks found.
c iw  work array .. see later comments.
C
*/
/**
	Initialise MatBool to be all false
	
	@param nr number of rows
	@param nc number of columns
*/
MatBool::MatBool(const unsigned &nr, const unsigned &nc){
	ip = VecInt(size_type(nr+1),0);
	lenr = VecInt(size_type(nr),0);
	icn = VecInt();
	ip[nr]=icn.end(); // must always equal this
}

/**
	Copy constructor
*/
MatBool::MatBool(const MatBool &a){
	ip = VecInt(a.ip);
	lenr = VecInt(a.lenr);
	icn = VecInt(a.icn);
}

/**
	Destructor
*/
MatBool::~MatBool(){
	// automatic
}

/**
	Assign a copy of this matrix to another
*/
MatBool &
MatBool::operator=(const MatBool &a){
	ip = VecInt(a.ip);
	lenr = VecInt(a.lenr);
	icn = VecInt(a.icn);
	return *this;
}

/**
	Element access operator

	@example A = MatBool(3,3); const bool A12 = A(1,2);
*/
const bool
MatBool::operator()(const unsigned &r, const unsigned &r){
	cerr <<"MatBool::("<<r<<","<<c<<"): ";
	unsigned &rowlen = lenr[r];
	if(rowlen==0){
		cerr << "false"<< endl; 
		return false;
	}
	VecInt::iterator i;
	VecInt::iterator first=&icn[0] + ip[r];
	VecInt::iterator last =&icn[0] + ip[r+1];
	for(i=first; i<last && *i < c; ++i){
		if(*i==c){
			cerr << "true" << endl;
			return true;
		}
	}

	cerr<< "false" << endl;
	return false;
}

/**
	Set an element of the MatBool (to true)
*/
void
MatBool::set(const unsigned &r, const unsigned &c){
	if((*this)(r,c)){
		cerr << "MatBool::set("<<r<<","<<c<"): already set" << endl;
		return;
	}
	VecInt::iterator j,first,last;

	// pointer to the row we're looking at
	first = &icn[0] + ip[r];
	last = &icn[0] + ip[r+1];

	// find the insertion point
	for(j=first; j<last && *j < c; ++j);

	icn.insert(j,c);
		
	// move the references in the IP array (inital positions for each row)
	for(unsigned i=r+1; i<n+1; ++i){
		ip[i]++;
	}
}

/**
	Clear an element of the MatBool (to false)
*/
void
MatBool::clear(const unsigned &r, const unsigned &c){
	if((*this)(r,c)){
		cerr << "MatBool::clear("<<r<<","<<c<<"): already clear" << endl;
	}

	VecInt::iterator j,first,last;
	first = &icn[0] + ip[r];
	last = &icn[0] + ip[r+1];

	// open up a space in icn
	for(j=first; j<last && *j < c; ++j);

	if(j==last)throw runtime_error("MatBool::clear("<<r<<","<<c<<"): unexpected condition j==last");

	icn.erase(j);
		
	// move the references in the IP array (inital positions for each row)
	for(unsigned i=r+1; i<n+1; ++i){
		ip[i]++;
	}
}

ostream &
operator<<(const ostream &s, const MatBool &a){
	unsigned r,c;
	cerr << "+";
	for(c=0; c<nc; ++c)cerr << "-";
	cerr << "+" << endl;
	for(r=0; r<nr; ++r){
		VecInt::iterator j,first,last;
		first = &icn[0] + ip[r];
		last = &icn[0] + ip[r+1];
		cerr << "|";
		for(c=0, j = first;c<nc;++c){
			while(j<last && *j < c)++j;
			if(*j==c)cerr<<"#";
			else cerr<<"+";
		}
		cerr << "|" << endl;
	}
	cerr << "+";
	for(c=0; c<nc; ++c)cerr << "-";
	cerr << "+" << endl;
}		
			
	
	

#ifndef MATBOOL_H
#define MATBOOL_H

/**
	Sparse Matrix of Booleans

	This is indended to reproduce the functionality of setup_ from the MC21A
	package, for compact storage of a sparse matrix of true/false values.

	This will be used to replace the FORTRAN_ARRAY(A,r,c,n) format currently
	used.
*/
class MatBool{

public:
	MatBool(const unsigned &nr, const unsigned &nc);
	~MatBool();

	SpMatBool(const SpMatBool &a);
	SpMatBool &operator=(const SpMatBool &a);

	inline const unsigned &rows() const{return nr};
	inline const unsigned &cols() const{return nc};

	const bool operator()(const unsigned &r, const unsigned &c);

private:
	unsigned nr;
	unsigned nc;
}

#endif
