/*

nla-ginac - Tarjan-blocking NLA solver with autodiff
Copyright (C) 2005  John Pye

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#include "nnes.h"
#include "nnessolver.h"

#include <vector>

using namespace std;

//------------------
// FANCY TRICK TO ALLOW NESTED CALLS TO NNES
// see http://www.experts-exchange.com/Programming/Programming_Languages/Cplusplus/Q_21505588.html

const unsigned NNESSolver::max_instances = 10;
unsigned NNESSolver::instance_count = 0;

NNESSolver *nnes_instance[NNESSolver::max_instances];

template<int i>
class NNESCallback{
public:
	static void getResFn(int*overfl, const unsigned*n1, double*res, const double*y){
		//cerr << "NNESCallback::getResFn: instance " << i << endl;
		//cerr << "NNESCallback::getResFn: y[0] = " << y[0] <<  endl;
		NNESSolver *nnessolver = nnes_instance[i];
		if(nnessolver==0){
			throw runtime_error("nnes_instance array contains null");
		}
		return nnessolver->ResFn(overfl,n1,res,y);
	}
	static void getJacobian(int *overfl, const unsigned *n1, double* J, const double*y){
		//cerr << "Calling Jacobian for instance " << i << endl;
		NNESSolver *s = nnes_instance[i];
		return s->Jacobian(overfl, n1, J, y);
	}
};

vector<NNES::ResFn *> NNESSolver::nnes_resfn_ptrs = vector<NNES::ResFn *>(max_instances);
vector<NNES::Jacobian *> NNESSolver::nnes_jacfn_ptrs = vector<NNES::Jacobian *>(max_instances);


#define NNES_INSTANCE(N) \
	nnes_resfn_ptr[N] = NNESCallback<N>::getResFn; \
	nnes_jacobian_ptr[N] = NNESCallback<N>::getJacobian;
//NNES_INSTANCE(0);
//NNES_INSTANCE(1);
//NNES_INSTANCE(2);
//NNES_INSTANCE(3);
//NNES_INSTANCE(4);

//---------------

NNESSolver::NNESSolver(const int &my_instance_number)
		: my_instance_number(my_instance_number)
{
	//cerr << "NNESSolver::NNESSolver:..."<<endl;
	if(instance_count >= max_instances){
		throw runtime_error("Too many NNES instances");
	}

	//cerr << "CREATED NNES SOLVER " << my_instance_number << endl;
	iopt["OUTPUT"]=0;
	iopt["QUIET"]=0;
    is_allocated=false;
}

NNESSolver::~NNESSolver(){
	if(is_allocated)freeMemory();
	instance_count--;
}

NNESSolver *
NNESSolver::Instance(){

	//cerr << "NNESSolver::Instance: instance_count = " << instance_count << endl;
	if(instance_count >= max_instances){
		throw runtime_error("Too many NNES instances (you can increase the number nnessolver.cpp)");
	}
	NNESSolver *nnes = new NNESSolver(instance_count);
	nnes_instance[instance_count] = nnes;
	instance_count++;
	return nnes;
}

const char *
NNESSolver::getType() const{
    return "NNESSolver";
}

void
NNESSolver::applyLowerBound(const unsigned &i, const double &b){
    //system->printdefaults();
    //cerr << "NNESSolver::applyLowerBound: lower bound: variable # " << i << ", '" << system->getVariableName(i) << "' = " << b << endl;

	boundl[i]=b;
}

void
NNESSolver::applyUpperBound(const unsigned &i, const double &b){
    //cerr << "NNESSolver::applyUpperBound: upper bound: variable # " << i << ", '" << system->getVariableName(i) << "' = " << b << endl;
	boundu[i]=b;
}


void
NNESSolver::visitSymbolic(SymbolicSystem *sys){
	this->visit(sys);
}

/**
	perform a system steady-state solution using NNES
*/
void
NNESSolver::visit(System *sys){
	throw runtime_error("NNESSolver::visit: can't use this now.");
	cerr << "NNESSolver:analyse:..." << endl;
}

void
NNESSolver::analyse(System *sys){
	cerr << "NNESSolver:analyse:..." << endl;
}

void
NNESSolver::solve(System *sys){
	cerr << "NNESSolver::solve: " << sys->getType() << endl;

	// Allocate space for vectors and arrays
	const unsigned &eqs = sys->getNumEqs();
	const unsigned &vars = sys->getNumVars();

    int trmcod;
    bool strack_alloc = false;

	try{
        // cerr << "NNES SOLVER... ENTERED TRY BLOCK" << endl;
		
		allocateMemory(sys);
		
		//cerr << "NNES ALLOCATED MEM" << endl;

        //nJacEvals=0;
		//nResEvals=0;

        iopt["MGLL"]=10;
		int nunit=6; // fortran file unit 6 = stdout
		char help[7];


		///  Call the fortran routine to set default values.

        //cerr << "NNES SOLVER... about to call setup" << endl;

		setup_(&iopt["ABSNEW"] ,&iopt["CAUCHY"] ,&iopt["DEUFLH"] ,&iopt["GEOMS"]  ,&iopt["LINESR"] ,
			&iopt["NEWTON"] ,&iopt["OVERCH"] ,&iopt["ACPTCR"] ,&iopt["ITSCLF"] ,&iopt["ITSCLX"] ,
			&iopt["JACTYP"] ,&iopt["JUPDM"]  ,&iopt["MAXEXP"] ,&iopt["MAXIT"]  ,&iopt["MAXNS"]  ,
			&iopt["MAXQNS"] ,&iopt["MINQNS"] ,&eqs            ,&iopt["NARMIJ"] ,&iopt["NIEJEV"] ,
			&iopt["NJACCH"] ,&iopt["OUTPUT"] ,&iopt["QNUPDM"] ,&iopt["STOPCR"] ,&iopt["SUPPRS"] ,
			&iopt["TRUPDM"]

			,&dopt["ALPHA"]  ,&dopt["CONFAC"] ,&dopt["DELTA"]  ,&dopt["DELFAC"] , &dopt["EPSMCH"]
			,&dopt["ETAFAC"] ,&dopt["FDTOLJ"] ,&dopt["FTOL"] ,&dopt["LAM0"]   ,&dopt["MSTPF"]
			,&dopt["NSTTOL"] ,&dopt["OMEGA"]  ,&dopt["RATIOF"] ,&dopt["SIGMA"]  ,
			&dopt["STPTOL"]

			,boundl , boundu , scalef , scalex , help
        );

		// Allow the system to specify solver options:

		//cerr << "OVERRIDING NNES DEFAULT OPTIONS..." << endl;
        //cerr << "This solver is type " << this->getType() << endl;
		//cerr << "FROM SYSTEM " << sys->getType() << endl;
		//cerr << "NNESSOLVER: THIS = " << this << endl;
        this->system=sys;
        //cerr << "NNESSolver::visit: asking system to apply solver options onto NNES." << endl;
		sys->applySolverOptions(this);
		sys->applyBounds(this);


		int original_stopcr = iopt["STOPCR"];

		///  Call the (fortran) nonlinear equation solver.

		//cerr << "Y VALUES FOR NNES ARE AT " << y << endl;
		/// Initial values

		if(vars != sys->getNumEqs()){
			stringstream ss;
			ss << "NNESSolver::visit: vars = " << vars << ", sys->getNumEqs() = " << sys->getNumEqs() << ", these should match for NNES." << endl;
			throw runtime_error(ss.str());
		}

		//cerr << "vars = " << vars << endl;
		//cerr << "eqs = " << eqs << endl;

		/*
		if(sys->getRel()==NULL){
			throw runtime_error("rel is not assigned");
		}
		*/

		for(unsigned i=0;i<vars;i++){
			y[i]=sys->getVariableDefault(i);
		}

		if(iopt["OUTPUT"]){
				cout << "INITIAL VALUES" << endl;
				sys->printvars(y);
				//cout << "CONSTANTS:" << endl;
				//sys->printconsts();
				cout << endl;
		}

        double whatever;

        double *orig_fvec=fvec;
        double *orig_fsave=fsave;

        double *strack = new double[iopt["MGLL"]];
        double *ftrack = new double[iopt["MGLL"]];
        strack_alloc=true;


		if(!iopt["QUIET"]){
			cerr << "STARTING NNES..." << endl << endl;
		}	
		//cerr << "Calling NNES..." <<endl;
		//cerr << "instance = " << my_instance_number << endl;
	
		/**
			@TODO: implement initialisation if NNESCallbacks using the trick at 
			http://osl.iu.edu/~tveldhui/papers/Template-Metaprograms/meta-art.html
		*/
		NNES::ResFn *resfnptr;
		NNES::Jacobian *jacfnptr;
		switch(my_instance_number){
			case 0: resfnptr = NNESCallback<0>::getResFn; jacfnptr = NNESCallback<0>::getJacobian; break;
			case 1: resfnptr = NNESCallback<1>::getResFn; jacfnptr = NNESCallback<1>::getJacobian; break;
			case 2: resfnptr = NNESCallback<2>::getResFn; jacfnptr = NNESCallback<2>::getJacobian; break;
			case 3: resfnptr = NNESCallback<3>::getResFn; jacfnptr = NNESCallback<3>::getJacobian; break;
			case 4: resfnptr = NNESCallback<4>::getResFn; jacfnptr = NNESCallback<4>::getJacobian; break;
			case 5: resfnptr = NNESCallback<5>::getResFn; jacfnptr = NNESCallback<5>::getJacobian; break;
			case 6: resfnptr = NNESCallback<6>::getResFn; jacfnptr = NNESCallback<6>::getJacobian; break;
			case 7: resfnptr = NNESCallback<7>::getResFn; jacfnptr = NNESCallback<7>::getJacobian; break;
			case 8: resfnptr = NNESCallback<8>::getResFn; jacfnptr = NNESCallback<8>::getJacobian; break;
			case 9: resfnptr = NNESCallback<9>::getResFn; jacfnptr = NNESCallback<9>::getJacobian; break;
			case 10: resfnptr = NNESCallback<10>::getResFn; jacfnptr = NNESCallback<10>::getJacobian; break;
			default: throw runtime_error("Invalid my_instance_number");
		}

		//cerr << "ResFn = " << resfnptr << endl;

		//int overfl1;
		//double *zzz= new double[eqs];

		//cerr << "Test of calling to ResFn:" << endl;
		//cerr << "y[0] = " << y[0] << endl;
		//cerr << "y = " << y << endl;
		//cerr << "zzz = " << zzz << endl;
		//(*resfnptr)(&overfl1, &eqs,zzz,y);
		//cerr << "Back from test" << endl;

		//cerr << "JacFn = " << jacfnptr << endl;

		iopt["OUTPUT"]=0;
		dopt["FDTOLJ"]=1e-3;

		nnes_(&iopt["ABSNEW"] ,&iopt["CAUCHY"] ,&iopt["DEUFLH"] ,&iopt["GEOMS"]  ,&iopt["LINESR"] ,// 5
			&iopt["NEWTON"] ,&iopt["OVERCH"] ,&iopt["ACPTCR"] ,&iopt["ITSCLF"] ,&iopt["ITSCLX"] ,// 10
			&iopt["JACTYP"] ,&iopt["JUPDM"]  ,&iopt["MAXEXP"] ,&iopt["MAXIT"]  ,&iopt["MAXNS"]  ,// 15
			&iopt["MAXQNS"] ,&iopt["MGLL"]   ,&iopt["MINQNS"] ,&eqs            ,&iopt["NARMIJ"] ,// 20
            &iopt["NIEJEV"] ,&iopt["NJACCH"] ,&whatever         ,&nunit           ,&iopt["OUTPUT"] ,// 25
            &iopt["QNUPDM"] ,&iopt["STOPCR"] ,&iopt["SUPPRS"] ,&trmcod, &iopt["TRUPDM"] // 30

            ,&dopt["ALPHA"]  ,&dopt["CONFAC"] ,&dopt["DELTA"] ,&dopt["DELFAC"] , &dopt["EPSMCH"] // 35
            ,&dopt["ETAFAC"] ,&dopt["FCNNEW"] ,&dopt["FDTOLJ"] ,&dopt["FTOL"]  ,&dopt["LAM0"]   // 40
            ,&dopt["MSTPF"] ,&dopt["NSTTOL"] ,&dopt["OMEGA"]  ,&dopt["RATIOF"],&dopt["SIGMA"]  , // 45

            &dopt["STPTOL"] ,a     , boundl , boundu , delf   , // 50
			fsave  , ftrack , fvec   , fvecc  , h      , // 55
			hhpi   , jac    , plee   , rdiag  , s      , // 60
			sbar   , scalef , scalex , sn     , ssdhat , // 65
			strack , vhat   , wv1    , wv2    , wv3    , // 70
			wv4    , y      , xplus  , xsave  , help // 75

			, resfnptr, jacfnptr // 77
		);

       // cerr << "RETURNED FROM NNES: " << trmcod << endl;

        if(fvec != orig_fvec){
            throw runtime_error("Fvec has changed");
        }

        if(fsave != orig_fsave){
            throw runtime_error("Fsave has changed");
        }

		const char *message;
		switch(trmcod){
			case 0:	 message = "NOT CONVERGED"; break;
			case 1:  message = "CONVERGED ON STEP SIZE (ONLY)"; break;
			case 2:  message = "CONVERGED ON FUNCTION VALUE"; break;
			case 3:  message = "CONVERGED ON STEP AND FUNC VAL (WITH BOTH REQD)"; break;
			case 12: message = "CONVERGED ON STEP AND FUNC VAL (WITH EITHER OK)"; break;
		}

		if((original_stopcr==2 || original_stopcr==3) && trmcod==1){
			message = "NOT CONVERGED (FUNC VALUE FAILED TO CONVERGE, STEP CONVERGED)";
            trmcod=0;
		}

		/// Calculate the final norm of the residuals and copy the result to X

		// res = zero;
		if (trmcod)
			for (unsigned i=0; i < eqs; i++)
				y[i] = xplus[i];

		// Perform a final residual evaluation to check the magnitude of the error:
        int overfl;
		sys->ResFn(&overfl, &eqs, fvec, y);

		double nnes_res;
		// Normalise:
		twonrm_(&overfl,&iopt["MAXEXP"],&eqs,&dopt["EPSMCH"],&nnes_res,fvec);

		// std::cout << "NNESStrategy::solve: res = " << nnes_res << ", trmcod = " << trmcod << std::endl;

		// int retcod = (trmcod == 2) ? 0 : -1;

		/// Report results

		//cout << "Return code from NNES is " << retcod << endl;

        //sys->printvars(y);

		char esc = unsigned(27);

		if(!iopt["QUIET"]){
			//cout << esc << "[1m" << "CONSTANTS:" << esc << "[0m" << endl;
			//dsys->printconsts();
			//cout << esc << "[1m"<< "VARIABLES:" << esc << "[0m"<< endl;
			//sys->printvars(y);
			sys->printresids(fvec);

		}

		// cout << esc << "[2m" << nJacEvals << " jacobian and " << nResEvals << " residual evaluations." << esc << "[0m" << endl;
		// return(retcod);

        if(trmcod==0){
            throw convergence_error(message);
        }

		//cerr << "setSolutionVector..." << endl;
		/// save solution
		sys->setSolutionVector(y);
		//cerr << "setSolutionVector...done" << endl;

        if(!iopt["QUIET"]){
			cerr << "PRINT3" << endl;
            sys->printsolution();
        }
		//cerr << "printsolution...done" << endl;

		if(!iopt["QUIET"]){
			cerr << endl << esc << "[1;32m"<< "NNES RETURNED: " << message << esc << "[0m" << endl << endl;
		}
		//cerr << "exit nnessolver...." << endl;
	}catch(convergence_error &e){
		cerr << endl << esc << "[1;31mCONVERGENCE FAILED at" << esc << "[0m" << endl;
		sys->printvars(y);
        throw;
    }catch(std::exception &e){
		std::stringstream s;
		s << "NNESSolver::visit: runtime_error: " << e.what();

		throw std::runtime_error(s.str());
	}
}

void
NNESSolver::allocateMemory(System *sys){
	//cerr << "NNESSolver::allocateMemory" << endl;
    unsigned eqs = sys->getNumEqs();
    unsigned vars = sys->getNumVars();

    //cerr << "NNESSolver::allocateMemory: eqs = " << eqs << endl;

	unsigned eqs2 = eqs*eqs;
	unsigned eqsvars = eqs*vars;

	unsigned bank_size = 4 * eqs2 + eqsvars + 20*eqs + 2*vars;

	if(is_allocated){
		if(bank_size > last_bank_size){
			cerr << "NNESSolver::allocateMemory: freeing instance " << my_instance_number << endl;
			freeMemory();
	    }else{
			// cerr << "NNESSolver::allocateMemory: reusing instance " << my_instance_number << endl;
			return;
		}
	}

	cerr << "NNESSolver::allocateMemory: allocated " << bank_size << " for instance "<< my_instance_number << endl;

	data_bank = new double[bank_size];
   	jac = data_bank; new double[eqs*eqs];
	a = jac + eqs2; //new double[eqs*eqs];
	h = a + eqs2; //new double[eqs * eqs];
	plee = h + eqs2;//new double[eqs * eqs];
    Jnum = plee + eqs2; // new double[eqs*vars];

	boundl = Jnum + eqsvars;
	boundu = boundl+eqs;
	delf = boundu+eqs;
	fsave = delf+eqs;
	fvec = fsave+eqs;

	fvecc = fvec+eqs;
	hhpi = fvecc+eqs;
	rdiag = hhpi+eqs;
	s = rdiag+eqs;
	sbar = s+eqs;

	sn = sbar+eqs;
	ssdhat = sn+eqs;
	vhat = ssdhat+eqs;
	wv1 = vhat+eqs;
	wv2 = wv1+eqs;

	wv3 = wv2+eqs;
	wv4 = wv3+eqs;
	xplus = wv4+eqs;
	xsave = xplus+eqs;
	scalef = xsave+eqs;

	y = scalef+eqs;
	scalex = y + vars;

    is_allocated=true;
	last_bank_size = bank_size;
}


void
NNESSolver::freeMemory(){
    if(!is_allocated){
        throw runtime_error("NNESSolver::freeMemory: memory is not currently allocated");
    }

	delete[] data_bank;
    is_allocated=false;

    //cerr << "Done deleting NNES data" << endl;
} // end solveSteady
