/*

nla-ginac - Tarjan-blocking NLA solver with autodiff
Copyright (C) 2005  John Pye

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#ifndef NNESSOLVER_H
#define NNESSOLVER_H

#include <stdexcept>
#include <sstream>
#include <vector>

#include "solver.h"
#include "nnes.h"

class NNESSolver : public Solver
{

public:
	static NNESSolver *Instance();

	virtual void visit(System *s);
	virtual void visitSymbolic(SymbolicSystem *s);
	virtual void analyse(System *s);
	virtual void solve(System *s);

    virtual const char *getType() const;
    
	/// Static methods that FORTRAN can access:
	static void staticResFn(int *overfl, const unsigned *n1, double* res, const double* y);
	static void staticJacobian(int *overfl, const unsigned *n1, double* Jnum, const double* y);

	inline void ResFn(int *overfl, const unsigned *n1, double* res, const double* y){
		//std::cerr << "NNESSolver::ResFn:..." << std::endl;
		//std::cerr << "NNESSolver::system = " << system << std::endl;
		//std::cerr << "NNESSolver::system.getType = " << system->getType() << std::endl;
        try{
			return this->system->ResFn(overfl,n1,res,y);
		}catch(std::exception &e){
			std::stringstream ss;
			ss << "NNESSolver::ResFn: instance " << my_instance_number << ": " << e.what();
			throw std::runtime_error(ss.str());
		}
    }

	inline void Jacobian(int *overfl, const unsigned *n1, double* Jnum, const double* y){
        return this->system->Jacobian(overfl,n1,Jnum,y);
    }

	void applyLowerBound(const unsigned &i, const double &b);
	void applyUpperBound(const unsigned &i, const double &b);

	virtual ~NNESSolver();

private:
	NNESSolver(const int &my_instance_number);

public:
	static const unsigned max_instances;

private:
	const unsigned  my_instance_number;
	static unsigned instance_count;
	static std::vector<NNES::ResFn *> nnes_resfn_ptrs;
	static std::vector<NNES::Jacobian *> nnes_jacfn_ptrs;

    void allocateMemory(System *s);
    void freeMemory();

	double *boundl;
	double *boundu;

    // memory locations required by NNES FORTRAN routine
    bool is_allocated;
    double *data_bank; // all allocated space made in one hit
	unsigned last_bank_size;

    double *jac,*a ,*delf ,*fsave ,*ftrack ,*fvec ,*fvecc ,*h, *hhpi ,*plee,*rdiag ,*s ,*sbar;
    double *scalef,*scalex ,*sn ,*ssdhat ,*strack, *vhat ,*wv1 ,*wv2 ,*wv3 ,*wv4 ,*xplus ,*xsave;

    double *Jnum;
    double *y;
};

#endif
