/*

nla-ginac - Tarjan-blocking NLA solver with autodiff
Copyright (C) 2005  John Pye

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#include "symbolicsystem.h"
#include "nnessolver.h"
#include "roulette.h"
#include "tarjansolver.h"
#include "tearingsolver.h"
#include "units.h"
#include "iapws-sat.h"
#include "colebrook-sym.h"
#include "viscosity-sym.h"

using namespace std;
using namespace GiNaC;

class PipeConstrainedSystem: public SymbolicSystem{

	virtual const char *getType() const{return "PipeConstrainedSystem";}

	virtual void defineSystem(){

		CONST(p1,60.0 * bar/MPa);
		VARIABLE(m1,0.01);
		CONST(T1,250+273.15);

		CONST(hpipe,10. / kilo);
		CONST(Tamb,300);

        CONST(eps,0.046e-3);
		CONST(L,10);
		CONST(D,0.05);

		CONST(p4,50.0 * bar/MPa);

		CONST(A1,Pi * D*D/4);
		CONST(Apipe, Pi*D*L);

		VARIABLE(rho1,802);
		VARIABLE(h1,water_h(T1,802));

		/// first node
		VARIABLE(mu1,1e-5);
		VARIABLE(Re1,7e7);
        VARIABLE(f1,0.02);
        VARIABLE(DP1, double(-10.0 * bar/ MPa));

		VARIABLE(rho2,0.0013);
		VARIABLE(h2, 2000);
		VARIABLE(T2, 180 + 273.15);
		VARIABLE(p2,56.0 * bar/MPa);
		VARIABLE(m2,0.052);

		/// second  node
		VARIABLE(mu2,1e-5);
		VARIABLE(Re2,7e7);
        VARIABLE(f2,0.02);
        VARIABLE(DP2, double(-10.0 * bar/ MPa));

		VARIABLE(rho3,0.003);
		VARIABLE(h3, 1900);
		VARIABLE(T3, 140 + 273.15);
		VARIABLE(p3,53.0 * bar/MPa);
		VARIABLE(m3,0.051);

		/// third node
		VARIABLE(mu3,1e-5);
		VARIABLE(Re3,7e7);
        VARIABLE(f3,0.02);
        VARIABLE(DP3, double(-10.0 * bar/ MPa));

		VARIABLE(rho4,0.001);
		VARIABLE(h4, 1800);
		VARIABLE(T4, 160 + 273.15);
		//VARIABLE(p4,50.0 * bar/MPa);
		VARIABLE(m4,0.05);

		LOWERB(mu2,0);
		LOWERB(mu3,0);

        LOWERB(h2,0);
        LOWERB(rho2,0);
        LOWERB(T2,0);

        LOWERB(h3,0);
        LOWERB(rho3,0);
        LOWERB(T3,0);

        LOWERB(h4,0);
        LOWERB(rho4,0);
        LOWERB(T4,0);

		lst ff; ff =
			/// Inlet properties
			p1 - water_p(T1,rho1)
			, h1 - water_h(T1,rho1)

			/// First segment
			, mu1 - mu_Trho(T1,rho1)
            , Re1 - rho1*m1*4/Pi/D/mu1
            , f1 - f_ReEd(Re1,eps/D)
            , DP1 + 0.5 * f1 * L / D * pow(m1,2) / pow(A1,2) / rho1

            , (p2 - p1) - DP1

			, m2 - m1
			, h2 - water_h(T2,rho2)

			, p2 - water_p(T2,rho2)
			, m1 / 2 * ( h2 - h1) - hpipe * Apipe * (T1 /2  - Tamb)

			/// NEXT SEGMENT
			, mu2 - mu_Trho(T2,rho2)
            , Re2 - rho2*m2*4/Pi/D/mu2
            , f2 - f_ReEd(Re2,eps/D)
            , DP2 + 0.5 * f2 * L / D * pow(m2,2) / pow(A1,2) / rho2

            , (p3 - p2) - DP2
			, m3 - m2
			, h3 - water_h(T3,rho3)

			, p3 - water_p(T3,rho3)
			, m2 / 2 * ( h3 - h2) - hpipe * Apipe * (T2 /2  - Tamb)

			/// NEXT SEGMENT

			, mu3 - mu_Trho(T3,rho3)
            , Re3 - rho3*m2*4/Pi/D/mu3
            , f3 - f_ReEd(Re3,eps/D)
            , DP3 + 0.5 * f3 * L / D * pow(m3,2) / pow(A1,2) / rho3

            , (p4 - p3) - DP3
			, m4 - m3
			, h4 - water_h(T4,rho4)

			, p4 - water_p(T4,rho4)
			, m3 / 2 * ( h4 - h3) - hpipe * Apipe * (T3 /2  - Tamb)

		;

        setExpressions(ff);

	}

	virtual void applySolverOptions(Solver *solver) const{
		//cerr << "COLEBROOK: SET SOLVER OPTIONS" << endl;
		// do bounds checking in NNES:

        //OPTION(ITSCLX,10);
        //OPTION(ITSCLF,9);
		OPTION(MAXIT,50);
        OPTION(STOPCR,2);
        OPTION(JACTYP,0);
        OPTION(FDTOLJ,8e-3);
        OPTION(OUTPUT,1);
        //OPTION(SHOW_BLOCKS,1);
	}

};

int
main(){
    PipeConstrainedSystem P;
	///Roulette *S = new Roulette();
	TarjanSolver *S = new TarjanSolver();

	//TarjanSolver *S = new TarjanSolver(TARJAN_TEARING);

	cerr << "main..." << endl;
    P.accept(S); // yes, i'll take a seat at the roulette table :)

	return 0;
}
