/*

nla-ginac - Tarjan-blocking NLA solver with autodiff
Copyright (C) 2005  John Pye

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#include <cstdlib>

#include "roulette.h"
#include "nnessolver.h"

using namespace std;

typedef struct{
	double lower;
	double upper;
} Bound;

Roulette::Roulette(){
	srand((unsigned)time(0));
}

void
Roulette::applyLowerBound(const unsigned &i, const double &b){
	boundl[i]=b;
}

void
Roulette::applyUpperBound(const unsigned &i, const double &b){
	boundu[i]=b;
}

void
Roulette::visit(System *sys){
	cerr << "LET'S ROLL..." << endl;

	char esc=unsigned(27);
	bool solved=false;
	NNESSolver *nnes = NNESSolver::Instance();
	nnes->setOption("QUIET",1);
	nnes->setOption("OUTPUT",0);
	nnes->setOption("CATCH",0);

	typedef map<int,Bound> BoundMap;

	BoundMap orig;
	BoundMap cvg;

	for(unsigned i=0;i < sys->getNumEqs();i++){
		if(sys->isBounded(i)){
			double l=sys->getLowerBound(i);
			double u=sys->getUpperBound(i);
			Bound b = {l,u};
			orig[i] = b;
		}
	}

	if(!orig.size()){
		throw runtime_error("Bounds have not been specified for any variables in roulette");
	}

	bool changed;
	while(!solved){
		try{
			BoundMap::const_iterator i;
			BoundMap::iterator j;
			for(i=orig.begin();i!=orig.end();i++){
				double l,u,v;

				l = (*i).second.lower;
				u = (*i).second.upper;
				if(l<=0 || u <= 0){
					// linear spread:
					v=l + (u-l)/(RAND_MAX) * rand();
				}else{
					// log spread:
					l = log(l);
					u = log(u);
					v=exp(l + (u-l)/(RAND_MAX) * rand());
				}

				//cerr << "Setting " << sys->getVariableName((*i).first) << " to " << v << endl;
				sys->setVariableDefault((*i).first,v);
			}

			sys->accept(nnes);

			changed=false;
			for(i=orig.begin();i!=orig.end();i++){
				double v = sys->getVariableDefault((*i).first);
				int n = (*i).first;
				if((j=cvg.find(n)) == cvg.end()){
					//changed=true;
					cvg[n].lower=v;
					cvg[n].upper=v;
				}else{
					if((*j).second.lower > v){
						//cerr << "Updating " << sys->getVariableName(n) << " lower limit to " << v << endl;
						(*j).second.lower=v;
						changed=true;
					}
					if((*j).second.upper < v){
						//cerr << "Updating " << sys->getVariableName(n) << " upper limit to " << v << endl;
						(*j).second.upper=v;
						changed=true;
					}
				}

			}

			// after update, display convergent range as it stands
			if(changed){
				cerr << endl;
				for(j=cvg.begin();j!=cvg.end();++j){
					int n = (*j).first;
					cerr << endl << esc << "[1;34m" << sys->getVariableName(n) << esc << "[0m: " << esc << "[0m" << cvg[n].lower << esc << "[2m - " << esc << "[0m" << cvg[n].upper;
				}
			}else{
				cerr << esc << "[2;32m." << esc << "[0m";
			}

		}catch(convergence_error &e){
			cerr << esc << "[2;31m." << esc << "[0m";
			// cerr << esc << "[2;31m" << e.what() << esc << "[0m" << endl;
			solved=false;
		}catch(exception &e){
			cerr << esc << "[1;31mError: " << e.what() << esc << "[0m" << endl;
			exit(1);
		}
	}
}

