/*

nla-ginac - Tarjan-blocking NLA solver with autodiff
Copyright (C) 2005  John Pye

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#include "iapws-low.h"
#include <iostream>
#include <stdexcept>
#include <sstream>

using namespace std;
using namespace GiNaC;


/**
	This program tests the saturation curve values from the 'revised supplementary relice on sautration properties of ordinary water substance, IAPWS 1992.
*/

#define VERBOSE

#undef assert
#ifdef VERBOSE
#define assert(EXPR) cerr<<"Assertion: " << #EXPR<< ": " << ((EXPR) ? "OK" : "FALSE") << endl;
#else
#define assert(EXPR) if(!(EXPR))throw runtime_error(#EXPR)
#endif

/**
	Get percent different between parameter values
*/
numeric PctDiff(const ex & a, const ex & b)
{
	if(!is_a<numeric>(a)){
		stringstream s;
		s << "PctDiff argument a = " << a << " is not numeric";
		throw runtime_error(s.str());
	}
	if(!is_a<numeric>(b)){
		stringstream s;
		s << "PctDiff argument b = " << b << " is not numeric";
		throw runtime_error(s.str());
	}

	numeric aa=ex_to<numeric>(a);
	numeric bb=ex_to<numeric>(b);

	if (aa==0 && bb==0){
		return 0.0;
	}

	return 100.0 * abs((aa-bb)/bb);
}

/**
	Test values from given inthe release on Table 1
*/
void TestTable1Points(){

    cerr << setprecision(6);

	numeric values[3][3] = {
		{ 273.16,	999.789, 0.00485426 }
		,{ 373.1243,	958.365, 0.597586 }
		,{ 647.096,	322,	322 }
	};

	for(int i = 0;i < 3; i++){
		numeric *row = values[i];

		ex T = row[0];
		ex rhof = row[1];
        ex rhof_calc = IAPWS95symbolic::rhof(T);

		//cerr << "rhof_calc = " << rhof_calc << ", rhof_ref = " << rhof << endl;
		assert(PctDiff(rhof_calc, rhof) < 0.001);

		ex rhog = row[2];
        ex rhog_calc = IAPWS95symbolic::rhog(T);

		//cerr << "rhog_calc = " << rhog_calc << ", rhog_ref = " << rhog << endl;
		assert(PctDiff(rhog_calc , rhog) < 0.001);
        
	}
}

int main(){
	try{
		TestTable1Points();
	}catch(exception &e){
		cerr << "ERROR: " << e.what() << endl;
		exit(1);
	}
	exit(0);
}

