/*

nla-ginac - Tarjan-blocking NLA solver with autodiff
Copyright (C) 2005  John Pye

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#ifndef STEAMEXPR_H
#define STEAMEXPR_H

#include <ginac/ginac.h>
#include "iapws-low.h"
#include <vector>
#include <iostream>

using namespace std;

class SteamExprBase{
protected:
	SteamExprBase(){}
public:
	virtual ~SteamExprBase(){}
	virtual GiNaC::ex get(const GiNaC::ex &T1, const GiNaC::ex &rho1) const = 0;
	virtual GiNaC::ex get_sat(const GiNaC::ex &T1, const GiNaC::ex &rho1) const = 0;
	virtual GiNaC::ex get_dT(const GiNaC::ex &T1, const GiNaC::ex &rho1) const = 0;
	virtual GiNaC::ex get_dT_sat(const GiNaC::ex &T1, const GiNaC::ex &rho1) const = 0;
	virtual GiNaC::ex get_drho(const GiNaC::ex &T1, const GiNaC::ex &rho1) const = 0;
	virtual GiNaC::ex get_drho_sat(const GiNaC::ex &T1, const GiNaC::ex &rho1) const = 0;
	virtual GiNaC::ex expr(const GiNaC::ex &T1, const GiNaC::ex &rho1) const = 0;
};


template<char property>
class SteamExpr : public SteamExprBase{

	private:

		GiNaC::symbol rho;
		GiNaC::symbol T;

		GiNaC::ex rhof, rhog;

		GiNaC::ex x, x_sat, xf,xg;
		GiNaC::ex x_dT, x_dT_sat;
		GiNaC::ex x_drho, x_drho_sat;

	private:
		SteamExpr(){

			rhof = IAPWS95symbolic::rhof(T);
			rhog = IAPWS95symbolic::rhog(T);

			x = expr(T,rho);
			xf = expr(T,rhof);
			xg = expr(T,rhog);

			x_dT = x.diff(T);
			x_drho = x.diff(rho);

			x_sat = xf + (rho - rhof)/(rhog - rhof) * (xg- xf);
			x_dT_sat = x_sat.diff(T);
			x_drho_sat = x_sat.diff(rho);
		}

	public:

		virtual ~SteamExpr(){}

		inline GiNaC::ex get(const GiNaC::ex &T1, const GiNaC::ex &rho1) const{
			//cerr << "RETURNING EXPR " << property << "..." << endl;
			GiNaC::ex ex1 = x.subs(T==T1).subs(rho==rho1);
			//cerr << "DONE with expr" << endl;
			return ex1;
		}

		inline GiNaC::ex get_sat(const GiNaC::ex &T1, const GiNaC::ex &rho1) const{
			//cerr << "RETURNING SAT " << property << endl;
			return x_sat.subs(T==T1).subs(rho==rho1);
		}

		inline GiNaC::ex get_dT(const GiNaC::ex &T1, const GiNaC::ex &rho1) const{
			//cerr << "RETURNING DT " << property << endl;
			return x_dT.subs(T==T1).subs(rho==rho1);
		}

		inline GiNaC::ex get_dT_sat(const GiNaC::ex &T1, const GiNaC::ex &rho1) const{
			//cerr << "RETURNING DT_SAT " << property << endl;
			return x_dT_sat.subs(T==T1).subs(rho==rho1);
		}

		inline GiNaC::ex get_drho(const GiNaC::ex &T1, const GiNaC::ex &rho1) const{
			//cerr << "RETURNING DRHO " << property << endl;
			return x_drho.subs(T==T1).subs(rho==rho1);
		}

		inline GiNaC::ex get_drho_sat(const GiNaC::ex &T1, const GiNaC::ex &rho1) const{
			//cerr << "RETURNING DRHO_SAT " << property << endl;
			return x_drho_sat.subs(T==T1).subs(rho==rho1);
		}

		// template function...
		GiNaC::ex expr(const GiNaC::ex &T1, const GiNaC::ex &rho1) const;

	public:
		static const SteamExpr *Instance(){
			static SteamExpr *instance;
			if(!instance){
				//cerr << "------- NEW STEAMEXPR " << property <<"  --------" << endl;
				instance = new SteamExpr();
			}
			return instance;
		}

};

#endif
