/*

nla-ginac - Tarjan-blocking NLA solver with autodiff
Copyright (C) 2005  John Pye

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#include "symbolicsystem.h"
#include "solver.h"

#include <ginac/ginac.h>
#include <iostream>
#include <sstream>
#include <stdexcept>

using namespace std;
using namespace GiNaC;


SymbolicSystem::SymbolicSystem() : System(){
	is_defined=false;
	// cerr << "NEW SYMBOLIC SYSTEM" << endl;
}

SymbolicSystem::~SymbolicSystem(){
    // delete symbols
	if(is_defined){
        delete J;
        delete rel;
        delete Jval;
    }
}

const char *
SymbolicSystem::getType() const{
	return "SymbolicSystem";
}

const bool
SymbolicSystem::isSymbolic() const{
	return true;
}

bool
symbol_pt_lt::operator()(const symbol *a, const symbol *b) const{
    if(a < b)return true;
    return false;
}

/// EQUATIONS (EXPRESSIONS)

const lst &
SymbolicSystem::getExpressions(void) const{
	return F;
}

void
SymbolicSystem::setExpressions(const lst &F){
    this->F = F;
	this->eqs=F.nops();
}

const ex
SymbolicSystem::getEx(const unsigned &i) const{
	if(i >= F.nops()){
		stringstream ss;
		ss << "SymbolicSystem::getEx: expression " << i << " has not been assigned.";
		throw runtime_error(ss.str());
	}
	return F[i];
}

const string
SymbolicSystem::getExpression(const unsigned &i) const{
	stringstream ss;
	ss << F[i];
	return ss.str();
}


void
SymbolicSystem::addExpressions(const lst &exprs){
	// cerr << "SymbolicSystem::addExpressions: ..." << endl;
	lst::const_iterator i;
	for(i=exprs.begin();i!=exprs.end();++i){
		this->F.append(*i);
	}
	this->eqs = F.nops();
}

/// VARIABLES

SymbolicVariable &
SymbolicSystem::addVariable(const char *name, const double &defval){

	//cerr << "Adding " << ex_to<numeric>(initval.evalf()).to_double() << " to list of defaults." << endl; 
	unsigned i = variables.size();
	SymbolicVariable *var = new SymbolicVariable(name);
	var->setDefVal(defval);
	variables.push_back(var);
	++vars;
	const symbol &s = var->getSymbol();
	symbolnumbers[&s] = i;
	return *var;
}

const symbol &
SymbolicSystem::addSymbol(const char *name, const double &defval){
	//cerr << "SymbolicSystem::addSymbol(name = "<<name << ", defval = " << defval << ")" << endl;
	const SymbolicVariable &v = addVariable(name,defval);
	return v.getSymbol();
}

const symbol &
SymbolicSystem::addSymbol(const char *name, const ex &defval){
	try{
        if(!is_a<numeric>(defval.evalf())){
            stringstream ss;
            ss << "Variable " << name << " not given a numerical initial value; current evaluates to: " << defval.evalf();

            throw runtime_error(ss.str());
        }

        if(!name){
            stringstream ss;
            ss << "SymbolicSystem::addVariable: adding unnamed variable, value is " << defval;
            throw runtime_error(ss.str());
        }

		double defval_double = ex_to<numeric>(defval.evalf()).to_double();

		return addSymbol(name,defval_double);

    }catch(exception &e){
        stringstream s;
        s << "SymbolicSystem::addVariable: " << e.what();
        throw runtime_error(s.str());
    }
}

void
SymbolicSystem::inheritVariable(SymbolicVariable *var ){
	variables.push_back(var);
	++vars;
}


SymbolicVariable *
SymbolicSystem::getSymbolVar(const symbol *s) const{
    SymbolIntMap::const_iterator i;
	if( ( i=symbolnumbers.find(s) ) == symbolnumbers.end() ){
        stringstream ss;
        ss << "Symbol not numbered: " << s;
        throw runtime_error(ss.str());
    }
    return getVar((*i).second);
}

const unsigned &
SymbolicSystem::getSymbolNum(const symbol *s) const{
    SymbolIntMap::const_iterator i;
	if( ( i=symbolnumbers.find(s) ) == symbolnumbers.end() ){
        stringstream ss;
        ss << "Symbol not numbered: " << s;
        throw runtime_error(ss.str());
    }
    return (*i).second;
}

SymbolicVariable *
SymbolicSystem::getVar(const unsigned &i) const{
	//cerr << "SymbolicSystem::getVar(" << i << "):... " << endl;
	return variables[i];
}


/// CONSTANTS

//#define CONST(X,V) double X=double(V);C[#X]=&X;

constant&
SymbolicSystem::addConstant(const char *name, ex val){
    try{
        try{
            val = val.evalf();
        }catch(exception &e){
            stringstream s;
            s << "Unable to evaluate value " << val;
            throw runtime_error(s.str());
        }

        if(!is_a<numeric>(val)){
            stringstream s;
            s << "Constant " << name << " not given a numerical initial value; currently evaluates to: " << val;
            throw runtime_error(s.str());
        }

        constant *s = new constant(name,ex_to<numeric>(val));
        CO.push_back(s);
        C[name] = ex_to<numeric>(val).to_double();

        constant& sr = *s;
        return sr;
    }catch(exception &e){
        stringstream s;
        s << "SymbolicSystem::addConstant: " << e.what();
        throw runtime_error(s.str());
    }
}

constant&
SymbolicSystem::addConstant(const char *name, const double initval){
	return addConstant(name,numeric(initval));
}


/// BOUNDS

void
SymbolicSystem::addExpression(const ex &expr1){
	this->F.append(expr1);
	this->eqs = F.nops();
}

bool
SymbolicSystem::isBoundedSymbol(const symbol *s){
	// check for valid symbol
	return getSymbolVar(s)->isBounded();
}

void
SymbolicSystem::setLowerBound(const symbol &s, const double &val){
	cerr << "Setting lower bound, " << s << " = " << val << endl;
	getSymbolVar(&s)->setLowerBound(val);
}

void
SymbolicSystem::setUpperBound(const symbol &s, const double &val){
	cerr << "Setting upper bound, " << s << " = " << val << endl;
	getSymbolVar(&s)->setUpperBound(val);
}

/// SOLUTION

void
SymbolicSystem::accept(Solver *solver){
	cerr << "SymbolicSystem::accept: system " << getType() << " system with solver at " << solver << endl;
	this->acceptAnalyse(solver);
	this->acceptSolve(solver);
}

void
SymbolicSystem::acceptAnalyse(Solver *solver){
	cerr << "SymbolicSystem::acceptAnalyse: type = " << getType() << endl;
	doDefineSystem(*solver);
	solver->analyseSymbolic(this);
}

void
SymbolicSystem::acceptSolve(Solver *solver){
	cerr << "SymbolicSystem::acceptSolve:..."<<endl;
	solver->solveSymbolic(this);
}

void
SymbolicSystem::setSolvedVariable(const symbol &sym, const double &val){
	SymbolIntMap::iterator si;
	if((si = symbolnumbers.find(&sym)) == symbolnumbers.end()){
		throw runtime_error("SymbolicSystem::setSolvedVariable: invalid symbol, not found");
	}
	
	const unsigned &i = (*si).second;
	cerr << "Setting variable " << i << " to " << val << endl;
	System::setSolvedVariable(i,val);
}

/**
	Prepare symbolic jacobian etc
*/
void
SymbolicSystem::doDefineSystem(const Solver &solver){
	//cerr << "SymbolicSystem::doDefineSystem: ..." << endl;
	const char esc=unsigned(27);
	if(!is_defined){
		try{
			//cerr << "SymbolicSystem::doDefineSystem: defining system..." << endl;
			this->defineSystem();
		
			printbounds();

			eqs=F.nops();

			if(vars > eqs){
				stringstream s;
				s << "Too many variables! (" << vars << " variables > " << eqs << " equations)";
				throw runtime_error(s.str());
			}else if(eqs > vars){
				stringstream s;
				s << "Too many equations! (" << eqs << " equations > " << vars << " variables)";
				throw runtime_error(s.str());
			}

			// CONNECTIONS
			/**
				@TODO would be good to eliminate the need for equations like "m1=m2"...trick is to remove elements from the V array using command like "equivalent(m1,m2)"
			*/

			/// Problem size

			lst::const_iterator r;

			/// Display problem

            if(!solver.getIntegerOption("QUIET")){
                cerr << "N VARS: " << vars << ", N EQS: " << eqs << endl;

			/*
			cout << endl << "Solving the problem:" << endl;
			for(r=F.begin();r!=F.end();++r){
				cout << "   " << (*r).evalf() << " = 0" <<  endl;
			}
			*/
			//printconsts();

                cerr << esc << "[2m" << "In variables:" << esc << "[0m";
                for(unsigned c=0;c<vars;c++){
                    cout << " " << getVar(c)->getName();
                }
                cout << endl << endl;
            }

			/// Calculate symbolic jacobian

			J = new matrix(eqs,vars);
			//cerr << "Created matrix J, with " << eqs << "rows, " << vars << " cols" << endl;

			rel = new bool[eqs*vars];

			for(unsigned c=0;c<unsigned(vars);c++){
				unsigned ri;
				lst::const_iterator r;
				for(
					ri = 0
						, r = F.begin()
					; r != F.end()
					; ++r
						, ++ri
				){
					ex Jrc = (*J)(ri,c) = r->diff(getVar(c)->getSymbol());

					if(Jrc == 0){
						//cerr << "Relationship: " << N[c] << " NOT  in " << *r << endl;
						FORTRAN_ARRAY(rel,ri,c,eqs) = false; // no relationship
					}else{
						//cerr << "Relationship: " << N[c] << " is in " << *r << endl;
						FORTRAN_ARRAY(rel,ri,c,eqs) = true;	// related
					}
				}
			}

			/// Symbolic determinant

#ifdef EVAL_J_DET
			cerr << "Calculating symbolic determinant... ";
			Jdet = J->determinant();
			cerr << Jdet << endl << endl;
#endif

			//cerr << *J << endl;

            if(!solver.getIntegerOption("QUIET") && solver.getIntegerOption("SHOW_SPARSITY")){
                /// Print sparsity pattern
				printsparsity(rel);
            }

			// Newton's method
			//cerr << "<<<<<<<<<<<<<<<<<<<<<<<<<<<< CREATED JVAL, eqs = " << eqs << endl;
			Jval = new matrix(eqs,vars);

			//cerr << "SYSTEM " << getType() << " IS NOW DEFINED" << endl;
			is_defined=true;


		}catch(exception &e){
			/*
			cerr << "SymbolicSystem variables:";
			for(int c=0;c<vars;c++){
				cerr << " " << N[c];
			}
			cerr << endl;
			*/
			stringstream ss;
			ss << "Error defining system: " << e.what();
			throw runtime_error(ss.str());
		}catch(...){
            cerr << "Some kind of error was thrown..." << endl;
            exit(1);
        }
	}else{
        //cerr << "system is already defined " << endl;
    }
}

/**
	Clear out temporary storage etc
*/
void
SymbolicSystem::undefineSystem(){
	delete J;
	delete rel;
	delete Jval;
	is_defined=false;
}

/**
	Evaluate the local jacobian & return it to FORTRAN

	@param Jnum		pointer to where the jacobian should be stored.
	@param y 		pointer to the current variables at which the jacobian should be evaluated.
*/
void
SymbolicSystem::Jacobian(int *overfl, const unsigned *n1, double* Jnum, const double* y){
	try{
        *overfl = 0;

        if(*n1 != vars || *n1 != eqs){
            cout << "Invalid n1 in Jacobian fn" << endl;
            exit(1);
        }


#ifdef EVAL_J_DET
        // Compute the determinant here...
        //cerr << "Jacobian determinant...";
        ex Jdetval = Jdet;
        for(unsigned c=0;c<vars;c++){
            symbol *s = V[c];
            Jdetval = Jdetval.subs(*s == y[c]);
        }
        cerr << endl<< "det(J) = " << Jdetval.evalf() << endl;
#endif

		//cerr<<esc<<"[36mJ" << esc << "[0m";
        //cerr << "Jacobian...";
		//cerr << "J.rows = " << (*J).rows() << ", J.cols = " << (*J).cols() << endl;
		if(!J)throw runtime_error("J is not initialised");
		if(!Jval)throw runtime_error("Jval is not initialised");

        // Compute the NUMERICAL Jacobian at y
        (*Jval) = (*J);
        for(unsigned c=0;c<vars;c++){
            //cout << ".";
            //cout.flush();
            for(unsigned r=0;r<eqs;r++){

                ex Jrc = (*J)(r,c);

                //int offset = eqs*c+r;

                double *Jnumrc = &FORTRAN_ARRAY(Jnum,r,c,eqs);

                if(false==FORTRAN_ARRAY(rel,r,c,eqs)){
                    *Jnumrc = 0;
                }else if(is_a<numeric>(Jrc)){
                    *Jnumrc = ex_to<numeric>(Jrc).to_double();
                }else{
                    // For this element of the jacobian, substitute all required unknowns
                    for(unsigned v=0; v<vars; v++){
                        const symbol &s = getVar(v)->getSymbol();
                        Jrc = Jrc.subs(s == y[v]);
                    }

                    (*Jval)(r,c) = Jrc;
                    ex Jrceval = Jrc.evalf();
                    if(!is_a<numeric>(Jrceval)){
                        stringstream ss;
                        ss << "SymbolicSystem::Jacobian: unable to evaluate " << Jrc << ": evaluated to " << Jrceval << ", not numeric.";
                        throw runtime_error(ss.str());
                    }
                    *Jnumrc = ex_to<numeric>(Jrc.evalf()).to_double();
                }
            }
        }
       // cout << endl;

		//cerr << "Jval.determinant() = " << Jval->determinant().evalf() << endl;
        nJacEvals++;

        /*cerr << "J =["<<endl;
        for(int r=0;r<eqs;r++){
            for(int c=0;c<vars;c++){
                cerr << setprecision(2) << setw(7) << *(FORTRAN_ARRAY(Jnum,r,c,eqs)) << " ";
            }
            cerr << endl;
        }
        cerr << "]" << endl;
		*/
    }catch(exception &e){
        stringstream ss;
        ss << "SymbolicSystem::Jacobian: " << e.what();
        throw runtime_error(ss.str());
    }

}

/**
	Evaluate the local system residuals and return them to FORTRAN
*/
void
SymbolicSystem::ResFn(int *overfl, const unsigned *n1, double* res, const double* y){
	//cerr << "SymbolicSystem::ResFn:..." << endl;
#ifdef TRACK_VAR_CHANGES
	static double *lasty;
	static bool firsteval;
	if(!lasty){
		lasty = new double[*n1];
		firsteval=true;
	}
#endif
	try{
#ifdef TRACK_VAR_CHANGES
		if(!firsteval){
			bool changed=false;
			for(unsigned i=0;i<*n1;i++){
				if((y[i]-lasty[i])/lasty[i] > 2e-8 ){
					cout << N[i] << ": " << lasty[i] << " --> " << y[i] << " (" << ((y[i]-lasty[i])/lasty[i]*100)<< "%)" <<endl;
					changed=true;
				}
			}

			if(!changed){
				cout << "(!) No variables have changed" << endl;
			}
		}
#endif
		//cout << "RESFN.y = " << y << endl;

		//cerr<<esc<<"[35mR" << esc << "[0m";

		// Compute new residual
		//cout << "RESIDUALS: VALS ARE AT " << y << ", var1 value is " << y[1] << endl;

		
		/*
		cout << "Computing residual" << endl;
		for(int i=0;i<vars;i++){
			cout << "y[" <<i <<"] = " << y[i] << endl;
		}
		*/
		

		double ssr=0;

        const lst &F = getExpressions();

		//cerr << "ABOUT TO LOOP THROUGH SUBS, eqs at " << &F << endl;
		//	cerr << "ABOUT TO LOOP THROUGH SUBS, eqs = " << F.evalf() << endl;

		unsigned ri;
		lst::const_iterator r;
		for(ri=0, r=F.begin(); r!=F.end(); ++r, ++ri){

			//cerr << " > EQUATION " << ri << endl;

			ex Bval = *r;

		    //cerr << "   = " << Bval << endl;

			for(unsigned c=0; c<vars; c++){
				//cerr << "    > VARIABLE " << c << endl;

				if(FORTRAN_ARRAY(rel,ri,c,eqs)){
					//cerr << "      RELATION EXISTS" << endl;
					try{
						const symbol &s=getVar(c)->getSymbol();
						//cerr << "      FOUND, " << s << endl;
						//cerr << "      substituting " << s << " = " << y[c] << " into " << Bval << endl;
						Bval = Bval.subs(s == y[c]);
						//cerr << "      OK" << endl;
					}catch(exception &e){
						stringstream ss;
						const symbol &s = getVar(c)->getSymbol();
                        cerr << "Error substituting " << s << " = "<< y[c] << " into " << Bval << ": " << e.what() << endl;
						ss<< "while substituting " << s << " = " << y[c] << " into " << Bval << ": " << e.what();
						throw runtime_error(ss.str());
					}
				}

				//cerr << "    <" << endl;
			}

            numeric Bnum;
            try{
                if(!is_a<numeric>(Bval.evalf())){
                    stringstream ss;
                    ss << "Not numeric: " << Bval.evalf();
                    throw runtime_error(ss.str());
                }
                Bnum = ex_to<numeric>(Bval.evalf());
            }catch(exception &e){
                stringstream ss;
                ss << e.what() << ": ";
                ss << "Unable to evaluate residual for equation " << ri << ": " << Bval << endl;
                throw runtime_error(ss.str());
            }


			double res1 = res[ri] = Bnum.to_double();

			ssr += res1*res1;
		}

		*overfl = 0;

#ifdef TRACK_VAR_CHANGES
		for(unsigned i=0;i<*n1;i++){
			lasty[i]=y[i];
		}
		firsteval=false;
#endif
	}catch(std::exception &e){
        stringstream ss;
        ss << "SymbolicSystem::ResFn: " << e.what();
        throw runtime_error(ss.str());
		//cerr << "ResFn error: " << e.what() << endl;
		//exit(1); // have to exit, since we're called from FORTRAN
	}

	nResEvals++;
	//cerr << "LEAVING RESFN" << endl;
}


/**
	This method is where you need to specify values of the NNES configuration rules.
	Consult the NNES reference manual (nnes.ps) at http://netlib.org/opt/
*/
void
SymbolicSystem::applySolverOptions(Solver *solver) const{
	// cerr << "SymbolicSystem::applySolverOptions" << endl;

	OPTION(JACTYP,0);
    OPTION(STOPCR,2);

}


void
SymbolicSystem::printresids(const double *res) const{
    cerr << esc << "[1m" << "RESIDUALS:" << esc << "[0m " << getType() << endl;
	for(unsigned i=0;i<eqs;i++){
        stringstream ss;
        ss << F[i];
		cout << esc << "[2m"<< "eq " << i << ": " << esc << "[0m"<< setw(10) << setprecision(4) << res[i] << " = " << ss.str() << endl;
	}
}

void
SymbolicSystem::printconsts() const{
	cerr << esc <<  "[2m" << CO.size() << " constants:" << esc << "[0m" << endl;
	//map<const char *,double>::iterator i;
	/*
	for(i=C.begin();i!=C.end();++i){
		cerr << (*i).first << " = " << ((*i).second) << endl;
	}
	*/
	for(vector<constant *>::const_iterator i=CO.begin();i<CO.end();++i){
		cerr << **i << " = " << ex(**i).evalf() << endl;
	}
}


void
SymbolicSystem::printsymbolnumbers() const{
    SymbolIntMap::const_iterator i;
    cerr << "There are " <<  symbolnumbers.size() << " symbolnumbers" << endl;
    for(i=symbolnumbers.begin();i!=symbolnumbers.end();++i){
        cout << (*i).first << ", " << *((*i).first) << " is symbol number " << (*i).second << endl;
    }
    cerr << "done symbol numbers :)" << endl;
}

void
SymbolicSystem::printequations() const{
	lst::const_iterator i;
    unsigned eqs = F.nops();
	cerr << esc << "[1mEQUATIONS:" << esc << "[2m (" << eqs << ")" << esc << "[0m" << endl;
	for(unsigned i=0;i<eqs;i++){
		stringstream ss;
		ss << F[i];
		cout << esc << "[2m"<< "eq " << i << ": " << esc << "[0m" << "0 = " << ss.str() << endl;
	}
}

void
SymbolicSystem::printsparsity(const bool *rel) const{
	
	cerr << esc << "[1m" << "SPARSITY PATTERN:" << esc << "[0m" << endl;
	cerr << "  " << setw(3) << "" << " +";
	for(unsigned c=0;c<vars;c++){
		cerr << "-";
	}
	cerr << "+" << endl;
	for(unsigned r=0;r<eqs;r++){
		cerr << esc << "[2m" << "eq" << setw(3) << r << esc << "[0m" << " |";
		for(unsigned c=0;c<vars;c++){
			if(FORTRAN_ARRAY(rel,r,c,eqs)) {
				cerr << esc << "[1;31m" << "#" << esc << "[0m";
			}else{
				cerr << esc << "[2m" << "+" << esc << "[0m";
			}
		}
		cerr << "|" << endl;
	}
	cerr << "  " << setw(3) << "" << " +";
	for(unsigned c=0;c<vars;c++){
		cerr << "-";
	}
	cerr << "+" << endl;
	cerr << endl;

}

void
SymbolicSystem::defineSystem(){
	cerr << "SymbolicSystem::defineSystem: nothing here (hopefully the equations have been defined elsewhere)";
}
