/*

nla-ginac - Tarjan-blocking NLA solver with autodiff
Copyright (C) 2005  John Pye

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#ifndef SYMBOLICSYSTEM_H
#define SYMBOLICSYSTEM_H

#include "system.h"
#include "exception.h"
#include "symbolicvariable.h"
#include "eqprocess.h"

#include <ginac/ginac.h>
#include <string>

// Assign a new symbol plus initial/default value
//#define G_SYMBOL(X,DEF) symbol X(#X); V.push_back(&X); D.push_back(DEF); N.push_back(#X);
#define VARIABLE(XXX,DEF) const symbol& XXX = this->addSymbol(#XXX,(DEF) )

// Assign a new constant
//#define G_CONST(X,V) double X=double(V);C[#X]=&X;
#define CONST(XXX,VAL) const constant& XXX = this->addConstant(#XXX,(VAL))

#define LOWERB(XXX,VAL) this->setLowerBound(XXX,VAL)
#define UPPERB(XXX,VAL) this->setUpperBound(XXX,VAL)

/**
    This is a comparator for use in map<symbol *,int>
    This mapping is required to facilitate setting of bounds using symbolic variables.
*/
class symbol_pt_lt{
public:
    bool operator()(const GiNaC::symbol *a, const GiNaC::symbol *b) const;
};

typedef std::map<const GiNaC::symbol*,unsigned,symbol_pt_lt> SymbolIntMap;
typedef std::map<const GiNaC::symbol*,double,symbol_pt_lt> SymbolDoubleMap;

typedef std::map<unsigned,std::set<unsigned> > TornEqsMap;

/**
	A system which is to be defined as a set of symbolic equations (using GiNaC)
	
	Any system that the 'user' defines will be a subclass of this one.
*/
class SymbolicSystem : public System{

public:

	typedef SymbolicVariable var_t;

	virtual const char *getType() const;
	virtual const bool isSymbolic() const;

	void solveSteady();

	virtual void applySolverOptions(Solver *solver) const;

	void undefineSystem();

	virtual void accept(Solver *s);
	virtual void acceptAnalyse(Solver *s);
	virtual void acceptSolve(Solver *s);

	virtual void ResFn(int *overfl, const unsigned *n1, double* res, const double* y);
	virtual void Jacobian(int *overfl, const unsigned *n1, double* Jnum, const double* x);

    const GiNaC::lst &getExpressions(void) const;
	const std::string getExpression(const unsigned &i) const;
	const GiNaC::ex getEx(const unsigned &i) const;

	const GiNaC::matrix *getJacobian() const;

	inline const unsigned &getNumEqs(void) const{return eqs;}
	inline const unsigned &getNumVars(void) const{return vars;}
	inline const std::vector<GiNaC::constant *> &getConstants(void) const{return CO;}

	virtual void printresids(const double *res) const;
	virtual void printconsts() const;
	virtual void printsymbolnumbers() const;
	virtual void printequations() const;
	virtual void printsparsity(const bool *rel) const;

    SymbolicVariable *getVar(const unsigned &i) const;
	SymbolicVariable *getSymbolVar(const GiNaC::symbol *sym) const;
	const unsigned &getSymbolNum(const GiNaC::symbol *sym) const;
	virtual void inheritVariable(SymbolicVariable *var);
	const std::vector<GiNaC::symbol *> &getVariables(void) const;
	const GiNaC::symbol *getVariable(const unsigned &i) const;

    void setSolvedVariable(const GiNaC::symbol &sym, const double &val);

protected:
	SymbolicSystem();
	virtual ~SymbolicSystem();

	virtual void doDefineSystem(const Solver &solver);
	virtual void defineSystem(void);

	SymbolicVariable &addVariable(const char *name, const double &defval);
	const GiNaC::symbol &addSymbol(const char *name, const GiNaC::ex &defval);
	const GiNaC::symbol &addSymbol(const char *name, const double &defval);

	bool isBoundedSymbol(const GiNaC::symbol *s);
	void setLowerBound(const GiNaC::symbol &s, const double &val);
	void setUpperBound(const GiNaC::symbol &s, const double &val);

	GiNaC::constant& addConstant(const char *name, GiNaC::ex val);
	GiNaC::constant& addConstant(const char *name, const double val);

    void setExpressions(const GiNaC::lst &F);
	void addExpressions(const GiNaC::lst &exprs);
	virtual void addExpression(const GiNaC::ex &expr);

protected:
	std::vector<GiNaC::constant *> CO;     ///< constants (vector of constant)
	SymbolIntMap symbolnumbers; ///< given a symbol, where in 'V' is its value held?

private:
	std::vector<SymbolicVariable *> variables;

	GiNaC::lst F; ///< the equations (F = 0), so F is just LHSs.

	std::map<const char *,double> C; ///< map of constant names to their values

	GiNaC::matrix *J; ///< symbolic jacobian
	GiNaC::matrix *Jval; ///< locally evaluated jacobian (temporary)

#ifdef EVAL_J_DET
	GiNaC::ex Jdet; ///< Symbolic jacobian determinant
#endif

};

#endif // SYMBOLICSYSTEM
