/*

nla-ginac - Tarjan-blocking NLA solver with autodiff
Copyright (C) 2005  John Pye

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#include "system.h"
#include "solver.h"

#include <ginac/ginac.h>
#include <iostream>
#include <sstream>

using namespace std;
using namespace GiNaC;

const char System::esc = unsigned(27);

System::System(){
	eqs=0;
	vars=0;
	is_defined=false;
	// nothing
}

System::~System(){
	// nothing
}


void
System::accept(Solver *solver){
	cerr << "System::accept: ..." << endl;
	acceptAnalyse(solver);
	acceptSolve(solver);
}

void
System::acceptSolve(Solver *solver){
	doDefineSystem(*solver);
	solver->solve(this);
}

void
System::acceptAnalyse(Solver *solver){
	cerr << "System::acceptAnalyse:..." << endl;
	doDefineSystem(*solver);
	solver->analyse(this);
}

/// TYPE INFORMATION

const char *
System::getType() const{
	return "System";
}

const bool
System::isSymbolic() const{
	return false;
}

/// VARIABLES AND DEFAULTS

const Variable *
System::getVar(const unsigned &i) const{
	Variable *var = variables[i];
	// cerr << "System::getVar: getting pointer to variable " << var->getName() << endl;
	return var;
}

const char *
System::getVariableName(const unsigned &i) const{
    //cerr << "getting variable name # " << i << " = " << (void *)(N[i]) << endl;
	return getVar(i)->getName();
}

const double &
System::getVariableDefault(const unsigned &i) const{
	return getVar(i)->getDefVal();
}

Variable &
System::addVariable(const char *name, const double &defval){
	//cerr << "Adding " << ex_to<numeric>(initval.evalf()).to_double() << " to list of defaults." << endl; 
	Variable *var = new Variable(name);
	var->setDefVal(defval);
	variables.push_back(var);
	++vars;
	return *var;
}

void
System::inheritVariable(Variable *var){
	variables.push_back(var);
	++vars;
}

/// BOUNDS

bool
System::isBounded(const unsigned &i) const{
	return getVar(i)->isBounded();
}

const double &
System::getLowerBound(const unsigned &i) const{
	cerr << "System::getLowerBound: " << i << endl;
	return getVar(i)->getLowerBound();
}

const double &
System::getUpperBound(const unsigned &i) const{
	cerr << "System::getUpperBound: " << i << endl;
	return getVar(i)->getUpperBound();
}

bool
System::isBoundedLower(const unsigned &i) const{
	return getVar(i)->isBoundedLower();
}

bool
System::isBoundedUpper(const unsigned &i) const{
	return getVar(i)->isBoundedUpper();
}

/// SOLVED VARIABLES

/// return the value which was solved

void
System::allocateSolution(){
	this->sol = vector<double>(vars);
}

const double &
System::getSolvedVariable(const unsigned &i) const{
	return sol.at(i);
}

void
System::setSolvedVariable(const unsigned &i, const double &val){
	if(sol.size()==0)allocateSolution();
	sol.at(i)=val;
}

void
System::setSolutionVector(const vector<double> &y){
	allocateSolution();
	vector<double>::const_iterator from;
	vector<double>::iterator to;
	for(from=y.begin(), to=sol.begin(); from < y.end() & to < sol.end(); ++from, ++to){
		*to=*from;
	}
}

void
System::setSolutionVector(const double *sol){
	//cerr << "SETTING SOLUTION... " << endl;
	//cerr << "vars = " << vars << endl;
	allocateSolution();
	for(unsigned i=0;i<vars;i++){
		//cerr << "sol[" << i << "] = " << sol[i] << endl;
		this->sol.at(i) = sol[i];
	}
	//cerr << "SOL.size = " << this->sol.size() << endl;
}

void
System::copySolution(const System *sys1){
	allocateSolution();
	for(unsigned i=0; i<vars; ++i){
		double yi = sys1->getSolvedVariable(i);
		this->sol.at(i) = yi;
	}
}


void
System::checkSolution(double *res){
	int overfl=0;
	ResFn(&overfl,&getNumEqs(),res,&sol[0]);
}

/// RELATIONSHIP MAP

const bool *
System::getRel() const{
	return rel;
}

/**
	Clear out temporary storage etc
*/
void
System::undefineSystem(){
	delete rel;
	is_defined=false;
}

const bool
System::isDefined() const{
	return is_defined;
}

/**
	This method is where you need to specify DEFAULT values of the NNES configuration rules.
	These will apply if you don't specify an applySolverOptions method in your own system spec.

	Consult the NNES reference manual (nnes.ps) at http://netlib.org/opt/
*/
void
System::applySolverOptions(Solver *solver) const{
	cerr << "System::applySolverOptions:..." << endl;
	// Set method (no way to get options in except by direct compiling)
	// output = 2;
	/*
	absnew=0;
	newton=0;
	linesr=0;
	*/
	//itsclx=1;
	//itsclf=1;
	//output=5;

    OPTION(JACTYP,0);
    OPTION(STOPCR,2);
	//jupdm=2;
	//ftol = 1e-10;
	cerr << "...done" << endl;
}

/**
    Apply any bounds on variables in this system to the solver.
*/
void
System::applyBounds(Solver *solver) const{
	//cerr << "System::applyBounds: ..."<< endl;
	bool found_bounds = false;
	for(unsigned i=0; i < vars; ++i){
		const Variable *var = getVar(i);
		if(var->isBoundedLower()){
			//cerr << "Lower bound: " << var->getName() << " = " << var->getLowerBound() << endl;
			solver->applyLowerBound(i,var->getLowerBound());
			found_bounds=true;
		}
		if(var->isBoundedUpper()){
			//cerr << "Upper bound: " << var->getName() << " = " << var->getUpperBound() << endl;
			solver->applyUpperBound(i,var->getUpperBound());
			found_bounds=true;
		}
	}
	if(!found_bounds){
		cerr << esc << "[2mNo bounds" << esc << "[0m" << endl;
	}

}


void
System::printvars(const double *y) const{
	for(unsigned i=0;i<vars;i++){
		cout << getVar(i)->getName() << " = " << y[i] << endl;
	}
}

void
System::printdefaults() const{
	for(unsigned i=0;i<vars;i++){
		const Variable *var = getVar(i);
		cout << var->getName() << " = " << var->getDefVal() << endl;
	}
}



void
System::printsolution() const{

	if(sol.size() != vars){
		throw runtime_error("System::printsolution: solution not found (sol.size()!=vars)");
	}
	
	cerr << esc << "[1mSOLUTION:" << esc << "[0m " << getType() << endl;
	for(unsigned i=0;i<vars;++i){
		cerr << getVar(i)->getName() << " = " << sol[i] << endl;
	}
	cerr << endl;
}

void
System::printbounds() const{
    stringstream ss;
	bool has_bounds = false;
	ss << esc << "[1m" << "BOUNDS:" << esc << "[0m" << endl;
	for(unsigned i=0; i<vars; ++i){
		const Variable *var = getVar(i);
		if(var->isBounded()){
			ss << esc << "[1m" << var->getName() << esc << "[0m:";
			if(var->isBoundedLower()) ss << " lower = " << var->getLowerBound();
			if(var->isBoundedLower() && var->isBoundedUpper()) ss << ",";
			if(var->isBoundedUpper()) ss << " upper = " << var->getUpperBound();
			ss << endl;
			has_bounds = true;
		}
	}
	if(has_bounds)cerr << ss.str();
	//if(!has_bounds){
	//	ss << "NO BOUNDS FOUND" << endl;
	//	cerr << ss.str();
	//}else{
	//	cerr << ss.str();
	//}
	
}

void
System::printresids(const double *res) const{
    cerr << esc << "[1m" << "RESIDUALS:" << esc << "[0m " << getType() << endl;
	for(unsigned i=0;i<eqs;i++){
        cout << esc << "[2m"<< "eq " << i << ": " << esc << "[0m"<< setw(10) << setprecision(4) << res[i] << endl;
	}
}
	
void
System::doDefineSystem(const Solver &s){
	throw runtime_error("System::doDefineSystem: needs to be overloaded");
}

void 
System::ResFn(int *overfl, const unsigned *n1, double* res, const double* y){
	throw runtime_error("System::ResFn: needs to be overloaded");
}

void
System::Jacobian(int *overfl, const unsigned *n1, double* Jnum, const double* x){
	throw runtime_error("System::Jacobian: needs to be overloaded");
}
