/*

nla-ginac - Tarjan-blocking NLA solver with autodiff
Copyright (C) 2005  John Pye

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#ifndef SYSTEM_H
#define SYSTEM_H

#include "exception.h"
#include "variable.h"

#include <iostream>
#include <vector>
#include <cmath>
#include <map>

// Access an 2-D array of double in the Fortran way...
#define FORTRAN_ARRAY(ARRAY,R,C,NR) ARRAY[(C)*NR +(R)]

#define OPTION(XXX,VAL) solver->setOption(#XXX,VAL);

typedef std::map<unsigned,double> BoundMap;

class Solver;

/**
	A generic system on which we will allow Solver objects to act.

	This class should contain all 'interface' requirements to allow
	iterative evalution of residuals and jacobian, but contain
	no mention of the underlying architecture used to
	evaluate the equations, etc (ie no GiNaC)
*/
class System {

public:
	virtual const bool isSymbolic() const;

	virtual const char *getType() const;

	virtual void accept(Solver *s);
	virtual void acceptAnalyse(Solver *s);
	virtual void acceptSolve(Solver *s);

	void solveSteady();

	virtual void applySolverOptions(Solver *solver) const;

	void undefineSystem();
	const bool isDefined() const;
 
	inline const unsigned &getNumEqs(void) const{return eqs;}
	inline const unsigned &getNumVars(void) const{return vars;}

	virtual const Variable *getVar(const unsigned &i) const;
	const char *getVariableName(const unsigned &i) const;
    const double &getVariableDefault(const unsigned &i) const;
	
	const bool *getRel() const;

	const double &getUpperBound(const unsigned &i) const;
	const double &getLowerBound(const unsigned &i) const;
	bool isBounded(const unsigned &i) const;
	bool isBoundedLower(const unsigned &i) const;
	bool isBoundedUpper(const unsigned &i) const;

	void allocateSolution();
	const double &getSolvedVariable(const unsigned &i) const;
	void setSolvedVariable(const unsigned &i, const double &val);
	void setSolutionVector(const std::vector<double> &sol);
	void setSolutionVector(const double *sol);
	void copySolution(const System *sys1);
	void checkSolution(double *res);

	virtual void ResFn(int *overfl, const unsigned *n1, double* res, const double* y);
	virtual void Jacobian(int *overfl, const unsigned *n1, double* Jnum, const double* x);

	virtual void printvars(const double *y) const;
	virtual void printdefaults() const;
	virtual void printbounds() const;
	virtual void printresids(const double *res) const;
	virtual void printsolution() const;

    void applyBounds(Solver *s) const;
	
protected:

	System();
	virtual ~System();

	virtual void doDefineSystem(const Solver &solver);
	virtual void defineSystem(void) = 0;


	Variable &addVariable(const char *name, const double &defaultval);
	virtual void inheritVariable(Variable *var);

protected:
	bool is_defined;

	std::vector<Variable *> variables;
	std::vector<double> sol; ///< solution values

	unsigned vars; ///< number of variables
	unsigned eqs; ///< number of equations

	unsigned nJacEvals; ///< number of evaluations of the jacobian
	unsigned nResEvals; ///< number of evaluations of the system residuals

	double *Jnum; ///< Fortran array jacobian, for sending to NNES. It's a dense array.

	bool *rel;

	static const char esc;
};

#endif
