/*

nla-ginac - Tarjan-blocking NLA solver with autodiff
Copyright (C) 2005  John Pye

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#include <iostream>
#include <ginac/ginac.h>

using namespace std;

#include "tarjansolver.h"
#include "tarjansystem.h"
#include "nnessolver.h"
#include "tearingsolver.h"

#include "eqprocess.h"

#include "printblocks.h"

#include "mc13d.h"
#include "mc21a.h"

#include <g2c.h>

char esc = unsigned(27);

TarjanSolver::TarjanSolver(){
	// options for this solver?
    iopt["SHOW_SPARSITY"]=1;
    iopt["BLOCK_OUTPUT"]=1;
}

TarjanSolver::~TarjanSolver(){
}

const char *
TarjanSolver::getType() const{
    return "TarjanSolver";
}

/// EQUATIONS

EqProcess &
TarjanSolver::getEqProcess(const unsigned &i){
	return equations[i];
}

const unsigned
TarjanSolver::getNumEqs() const{
	return equations.size();
}

/// BOUNDS

void
TarjanSolver::applyLowerBound(const unsigned &i, const double &b){
	cerr << "Applying bound in Tarjan solver";
	boundl[i]=b;
}

void
TarjanSolver::applyUpperBound(const unsigned &i, const double &b){
	cerr << "Applying bound in Tarjan solver";
	boundu[i]=b;
}

/**
	Visit the system: divide into blocks then solve each block
*/
void
TarjanSolver::visitSymbolic(SymbolicSystem *sys){
	//cerr << "TarjanSolver::visitSymbolic:..." << endl;

	cerr << "SYSTEM BLOCK ANALYSIS" << endl;
	this->analyseSymbolic(sys);

	// cerr << "SYSTEM BLOCK SOLUTION" << endl;
	this->solveSymbolic(sys);

}

/**
	@TODO: this should be optimised so that substitutions are only attempted where required
*/
void
TarjanSolver::eliminateVariable(const SymbolicVariable *var, const double &val){

	vector<EqProcess>::iterator ei;
	for(ei=equations.begin(); ei!=equations.end(); ++ei){
		const GiNaC::symbol &sym=var->getSymbol();
		ei->subs(sym==val);
	}

}

SymbolicSystem *
TarjanSolver::getBaseSystem(){
	return system;
}

/**
	Analyse the base system: permute to non-zero diagonal form if
	necessary; convert to block form with Tarjan's algorithm, and
	and create TarjanSolvers that will later be used to perform the solving.

	Also create local copies of the base system equations (as EqProcess
	objects) so that the progressive elminations of variables in the 
	system can be tracked.
*/
void
TarjanSolver::analyseSymbolic(SymbolicSystem *sys){
	cerr << "TarjanSolver::analyseSymbolic:..." << endl;
	this->system=sys;

    unsigned n = sys->getNumEqs();
    logical *relf;
    relf = new logical[n*n];
    logical *relf2;
    relf2 = new logical[n*n];

	try{

		cerr << "Beginning TarjanSolver block analysis..." << endl;

		// Extract the sparsity matrix:
		const bool *rel = sys->getRel();

		cerr << "Translating sparsity matrix to sparse form..." << endl;
		// translate to fortran TRUE_ and FALSE_ (g2c)
		unsigned nonzeros = 0;
		for(unsigned i=0;i<n;i++){
			for(unsigned j=0;j<n;j++){
				if(FORTRAN_ARRAY(rel,i,j,n)){
					FORTRAN_ARRAY(relf,i,j,n) = TRUE_;
					nonzeros++;
				}else{
					FORTRAN_ARRAY(relf,i,j,n) = FALSE_;
				}
			}
		}

		// convert to sparse format
		unsigned licn = 1000; // ? big enough?
		int *icn;
		icn = new int[licn];

		int *ip;
		ip = new int[n];

		int *lenr;
		lenr = new int[n];

		unsigned mm = n;

		/// convert to sparse: ip, icn, lenr.
		setupmc13d_(&n, relf, &mm, ip, icn, &licn, lenr);

		/// permute for non-zeros on the diagonal
		cerr << "Permuting rows for non-zeros on the diagonal..." << endl;
		unsigned *iperm;
		iperm = new unsigned[n];
		int *iw_21a;
		iw_21a = new int[4*n];
		int numnz;
		mc21a_(&n, icn, &licn, ip, lenr, iperm, &numnz, iw_21a);

		/*
		cerr << "PERMUTATION FOR NON-ZERO DIAGONAL:" << endl;
		for(int i=0; i<n; i++){
			cerr << "iperm[" << i << "] = " << iperm[i]-1 << endl;
		}
		*/

		/// permute the rows in the sparse format...

		int *ip1 = new int[n];
		int *lenr1 = new int[n];
		int *mrepi = new int[n]; // iperm backwards :)

		for(unsigned i=0; i<n; i++){
			ip1[i] = ip[iperm[i]-1];
			lenr1[i] = lenr[iperm[i]-1];
			mrepi[iperm[i]-1] = i-1;
		}

		if(!iopt["QUIET"]){
			cerr << "PERMUTED ROWS..." << endl;
			for(unsigned i=0;i<n;i++){
				cerr << esc << "[2meq" << setw(3) << iperm[i]-1 << ": " << esc << "[0m";
				int i_orig=iperm[i]-1;
				for(unsigned j=0;j<n;j++){
					if(FORTRAN_ARRAY(relf,i_orig,j,n)){
						cerr << esc << "[1;32m#" << esc << "[0m";
					}else{
						cerr << esc << "[2m." << esc << "[0m";
					}
				}
				cerr << endl;
			}
			cerr << endl;
		}

		// ensure diagonal elements are nonzero:
		for(unsigned i=0;i<n;i++){
			if(!FORTRAN_ARRAY(rel,iperm[i]-1,i,n)){
				stringstream ss;
				ss << "Sparsity matrix has non-zero on diagonal in row " << (i+1);
				throw runtime_error(ss.str());
			}
		}

		/// blocking analysis will be done on the permuted matrix, careful to remember!
		// allocate arrays for blocking analysis
		unsigned *ior; // index of original rows
		ior = new unsigned[n];

		unsigned num;

		int *ib; // index to rows for each block (up to n blocks)s
		ib = new int[n+1];

		int *iw; // work array... guessing that this will be big enough :(
		iw = new int[3*n];

		cerr << "Performing blocking analysis..." << endl;

		/// perform blocking analysis
		mc13d_(&n, icn, &licn, ip1, lenr1, ior, ib, &num, iw);

        MC13D::printblocks_perm(n,num,ib,iperm,ior,relf);
		ib[num] = n + 1;

		/*
		cerr << "Rows and columns re-mapped " << endl;
		for (unsigned i=0; i<n; ++i) {
			cerr << "R,C" <<  ior[i]-1 << " -> " << i << endl;
		}
		cerr << endl;
		*/

		/// keep a local list of expressions

		// cerr << "Retrieving local copy of equations from the base system... " << endl;

		for(unsigned i=0; i<n; ++i){
			SymbolicVariable *var = sys->getVar(ior[i]-1);
			const GiNaC::ex eq = sys->getEx(iperm[ior[i]-1]-1);
			equations.push_back(EqProcess(var, &eq));
		}

		/// create TarjanSystems (sub-systems) that a sub-solver can visit:
		
		bool have_onevar = false;
		for (unsigned i = 0; i < num; ++i) {
			cerr << "Creating block " << i << ", variables: ";
			if(ib[i+1] - ib[i] == 1)have_onevar = true;
			subsystems.push_back(TarjanSystem(this,ib[i]-1,ib[i+1]-1));
		}

		/// create sub-solver for each sub-system
		NNESSolver *nnes;
		if(have_onevar)nnes = NNESSolver::Instance();
		for(unsigned i = 0; i < num; ++i){
			cerr  << esc << "[1;34mAnalysing block " << i << esc << "[0m" << endl;
			Solver *subsolver;
			TarjanSystem &T = subsystems[i];
			if(T.getNumEqs() == 1){
				subsolver = nnes;
			}else{
				subsolver = new TearingSolver();
			}
			cerr << "Analysing block " << i << " ..." << endl;
			T.acceptAnalyse(subsolver);
			// subsolver->analyseSymbolic(&T);
			subsolvers.push_back(subsolver);
		}

		cerr << esc << "[1;34mTARJAN SOLVER ANALYSIS COMPLETE" << esc << "[0m" << endl;

	}catch(exception &e){
		stringstream ss;
		ss << "TarjanSolver::analyse: " << e.what();
		throw runtime_error(ss.str());
	}
}

/**
	Take the blocked system and solve for the blocks in order.
	Substitute the solution from earlier blocks into all the later equations.
	Report the solution to the system being solved.
*/
void
TarjanSolver::solveSymbolic(SymbolicSystem *sys){
	cerr << "TarjanSolver::solveSymbolic:..."<<endl;

    unsigned n = equations.size();
    double *tarjan_residuals = new double[n];

	// cerr << "TarjanSolver::solve:..." << endl;

	try{
		// this will point to the sub-system we are currently working on

		if(iopt["BLOCK_OUTPUT"]){
            cerr << endl << esc << "[1mSOLVING BLOCKS" << esc << "[0m" << endl;
        }

		cerr << "There are " << subsystems.size() << " subsystems" << endl;

		vector<TarjanSystem>::iterator si;
		vector<Solver *>::iterator vi;
		unsigned i;

		for (
				si=subsystems.begin(), i=0, vi=subsolvers.begin();
				si!=subsystems.end();
				++si, ++i, ++vi
		){
			
			TarjanSystem &subsystem = *si;
			Solver *subsolver = *vi;

			if(iopt["BLOCK_OUTPUT"]){
                cerr  << esc << "[1;34mBLOCK " << i << esc << "[0m " << subsystem.getType() << endl;
            }
			
			subsystem.getSolverEquations();
			subsystem.printequations();

			cerr << "Asking solver to come to " << subsystem.getType() << "..."<<endl;
			subsystem.acceptSolve(subsolver);
			cerr << "PRINT1" <<endl;
			si->printsolution();
		}

        cerr << "Tarjan solver completed." << endl<< endl;

		for(unsigned i=0; i<n; ++i){
			EqProcess &eq = equations[i];
			cerr << "i = " << i << ": " << eq.getBaseVariable()->getName() << " = " << eq.getSolution() << endl;
			sys->setSolvedVariable(eq.getBaseVariable()->getSymbol(),eq.getSolution());
		}
		
		// send the solution values back to the original system
		cerr << "PRINT2" << endl;
        sys->printsolution();

		cerr << "Checking system residuals directly using found solution:" << endl;
		sys->checkSolution(tarjan_residuals);
		cerr << "Printing residuals..." << endl;

        sys->printresids(tarjan_residuals);

		cerr << endl;
	}catch(exception &e){
		stringstream ss;
		ss << "TarjanSolver::solve: " << e.what();
		throw runtime_error(ss.str());
	}
}

void
TarjanSolver::printbase() const{
	cerr << "BASE EQUATIONS:" << endl;
	for(unsigned i=0;i<equations.size();i++){
		stringstream ss;
		ss << equations[i].getBaseEquation();
		cerr << esc << "[2m"<< "eq " << i << ": " << esc << "[0m"<< setw(10) << 0 << " = " << ss.str() << endl;
	}
	cerr << endl;
}

