/*

nla-ginac - Tarjan-blocking NLA solver with autodiff
Copyright (C) 2005  John Pye

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#include "tarjansolver.h"
#include "tarjansystem.h"
#include "symbolicsystem.h"
#include "solver.h"

using namespace GiNaC;
using namespace std;

/**
	Create new tarjan block sub-system

	@param base system of which this block is a part
	@param ior	index to original rows/columns
	@param begin begin index for variables in this sub-system
	@param end	end index for variables in this sub-substem ("end+1")
*/
TarjanSystem::TarjanSystem(TarjanSolver *tarjansolver, const unsigned &begin, const unsigned &end){
	this->tarjansolver = tarjansolver;

	// add the equations and variables from the tarjansolver's EqProcess array
	for(unsigned i=begin;i<end;i++){
		EqProcess &eq = tarjansolver->getEqProcess(i);
		cerr << eq.getBaseVariable()->getName() << " ";
		inheritVariable(eq.getBaseVariable());
		baseequations.push_back(&eq);
	}
	cerr << endl;
	eqs = baseequations.size();
}

TarjanSystem::~TarjanSystem(){
    // cerr << "DELETING TARJAN SYSTEM" << endl;
}

void 
TarjanSystem::acceptAnalyse(Solver *solver){
	cerr << "TarjanSystem::acceptAnalyse: ..." << endl;
	getSolverEquations();
	printequations();
	doDefineSystem(*solver);
	cerr << "TarjanSystem::acceptAnalyse: about to visit... "<< endl;
	solver->analyseSymbolic(this);
}

void
TarjanSystem::acceptSolve(Solver *solver){
	cerr << "TarjanSystem::acceptSolve:..." << endl;
	solver->solveSymbolic(this);
	cerr << "Notifying solved variables to base system..."<< endl;
	notifySolution();
	cerr << "TarjanSystem:: done acceptSolve " << endl;
}

/**
	Notify the TarjanSolver of the values of the solved variables
*/
void
TarjanSystem::notifySolution(){
	cerr << "TarjanSystem::notifySolution" << endl;
	vector<EqProcess *>::iterator ei;
	unsigned i;
	for(ei=baseequations.begin(), i=0; ei!=baseequations.end(); ++ei, ++i){
		//cerr << "TarjanSystem::notifySolution: " << (*ei)->getBaseVariable()->getName() << " = " << sol[i] << endl;
		tarjansolver->eliminateVariable((*ei)->getBaseVariable(),sol[i]);
		(*ei)->setSolution(sol[i]);
	}
}

const char *
TarjanSystem::getType() const{
	return "TarjanSystem";
}

void
TarjanSystem::getSolverEquations(){
	GiNaC::lst F;
	for(unsigned i=0; i<eqs; ++i){
		EqProcess *eq = baseequations[i];
		//cerr << "TarjanSystem::getSolverEquations: fetched equation " << eq->getCurrentEquation() << endl;
		F.append(eq->getCurrentEquation());
	}
	setExpressions(F);
}

void
TarjanSystem::defineSystem(){
	cerr << "TarjanSystem::defineSystem: get solver equations..." << endl;
	getSolverEquations();
	//cerr << "DEFINING SYSTEM" << endl;
}

void
TarjanSystem::doDefineSystem(Solver &solver){
	cerr << "TarjanSystem::doDefineSystem: what to do here?" << endl;
	SymbolicSystem::doDefineSystem(solver);
	if(!isDefined()){
		throw runtime_error("TarjanSystem::doDefineSystem: System still isn't defined");
	}
}

/**
	This will be called by NNESSolver when solving the TarjanSystem.

    The TarjanSystem has getSystem() =  the base system
    and the base system is being solved by the TarjanSolver.

    The base system can apply solver options onto the NNESSolver
    but to get at the base system, we need to use this->getSystem

    The base system will apply bounds on the NNES solver if required.
*/
void
TarjanSystem::applySolverOptions(Solver *solver) const{
	//cerr << "TarjanSystem::applySolverOptions:..." << endl;
	solver->setOption("QUIET",1);
	this->tarjansolver->getBaseSystem()->applySolverOptions(solver); // ??

	//cerr << "TARJANSYSTEM: SET SOLVER OPTIONS" << endl;

    //cerr << "TarjanSystem::setSolverOptions: solver type = " << solver->getType() << endl;
	//const System *base_system = this->getBaseSystem();
	//cerr << "TarjanSystem::getSolverOptions: base_system = " << base_system << endl;
	//cerr << "TarjanSystem::getSOlverOptions: getType = " << base_system->getType() << endl;
	//base_system->applySolverOptions(solver);

	solver->setOption("QUIET",1);
	solver->setOption("OUTPUT",0);
}

