/*

nla-ginac - Tarjan-blocking NLA solver with autodiff
Copyright (C) 2005  John Pye

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#include <iostream>
#include <ginac/ginac.h>
#include <map>
#include <cstring>

using namespace std;

#include "tearingsolver.h"
#include "nnessolver.h"
#include "tearingsystem.h"
#include "tornsystem.h"
#include "steward.h"
#include "tearsets.h"
#include "choosetearset.h"
#include "mc13d.h"
#include "mc21a.h"
#include "printblocks.h"
#include "tarjansolver.h"


TearingSolver::TearingSolver(){
	// options for this solver?
	iopt["SHOW_SPARSITY"]=1;
}

void
TearingSolver::applyLowerBound(const unsigned &i, const double &b){
	cerr << "Applying bound in tearing solver";
	//boundl[i]=b;
}

void
TearingSolver::applyUpperBound(const unsigned &i, const double &b){
	cerr << "Applying bound in tearing solver";
	//boundu[i]=b;
}

const char *
TearingSolver::getType() const{
    return "TearingSolver";
}

EqProcess &
TearingSolver::getEqProcess(const unsigned &i){
	return equations[i];
}

/**
	Visit the system: recursively analyse the block structure of the matrix
	then take a local copy of the equations and proceed through the recursive
	structure to solve for all equations.
*/
void
TearingSolver::visitSymbolic(SymbolicSystem *sys){

	// sys->printequations();

	cerr << esc << "[1mTearingSolver::visitSymbolic: analysing system..." << esc << "[0m" << endl;

	analyseSymbolic(sys);

	cerr << esc << "[1mTearingSolver::visitSymbolic: preparing subsystem objects" << esc << "[0m" << endl;

	// each TearingSolver knows where its equations are in the main system
	// it must solve for its tearing variables and return them to the parent solver...
	solveSymbolic(sys);
	
	cerr << esc << "[1mTearingSolver::visitSymbolic: ...done"  << esc << "[0m" << endl;

	// throw runtime_error("Tearing solver not yet completed");
}

/**
	Look for tearing loops in a given system of equations, already placed in nonzero
	diagonal format.

	Tearing is done using Steward's loop identificatio
	then a homebaked tearset chooser (simply the first minimal tearset found at the moment).
*/
void
TearingSolver::analyseSymbolic( SymbolicSystem *sys ){
	cerr << "TearingSolver::analyseSymbolic: type = " << sys->getType() << endl;

	try{
		if(!sys->isDefined()){
			stringstream ss;
			ss << sys->getType() << " is not yet defined (doDefineSystem)";
			throw runtime_error(ss.str());
		}

		const bool *rel = sys->getRel();
		MC13D::printrelmap(rel, sys->getNumEqs(), sys->getNumVars());
	
		if( strcmp(sys->getType(), "TarjanSystem") != 0 ){
			stringstream ss;
			ss << "TearingSolver::analyse(" << sys->getType() << "): unable to solve this type of system";
			throw runtime_error(ss.str());
		}
	
		unsigned n = sys->getNumEqs();
		if(n>1){
			// Check the matrix diagonal again
			for(unsigned r=0; r<n; ++r){
				if(!FORTRAN_ARRAY(rel,r,r,n)){
					throw runtime_error("Matrix contains a zero on the diagonal, can't proceed with tearing analysis");
				}
			}
	
			/// Find a suitable tearset
	
			// Identify loops in the relationship matrix
			CycleSet loops = steward(n,rel);
	
			if(loops.size()){
	
				// Find all complete tearsets
				vector<TearSet> ts = tearsets(loops, n);
				vector<TearSet>::iterator ti;
	
				cerr << "Found " << ts.size() << " possible tearsets." << endl;
	
				// Choose a 'best' tearset (needs work):
				TearSet tsmin = choosetearset(ts,*sys);
	
				cerr << "Torn variables: ";
				TearSet::iterator tsi;
				for(tsi=tsmin.begin(); tsi!=tsmin.end(); ++tsi){
					cerr << sys->getVariableName(*tsi) << " ";
				}
				cerr << endl;
	
				cerr << "Retrieving local copy of equations from the base system... " << endl;
				
				for(unsigned i=0; i<n; ++i){
					SymbolicVariable *var = sys->getVar(i);
					//cerr << "Got var at " << var;
					//cerr << ", symbol is " << var->getSymbol() << endl;
					const GiNaC::ex eq = sys->getEx(i);
					equations.push_back(EqProcess(var, &eq));
				}
	
				/// Eliminate the torn variables
	
				cerr << "Eliminating torn variables..." << endl;
	

				/**
					@TODO
					The torn system has all of the variables from the current system
					but some equations need to be flagged as having no relationship
					to the torn variables.

					This can be done by eliminating those variables from the main system equations
					but such an approach is doomed since in some cases, substituting a particular
					variable's value could eliminate more than one variable: eg

						(1 - x) / (y + z) = 0

					If you substitute x = 1 to the above, it will become 0 = 0, and the
					relationship with y and z will be lost.

					Hence, the only valid way to do this is to take the system symbolic jacobian
					inspect the matrix and remove those components of the jacobian which contain
					reference to the eliminate variable, and then use the revised matrix
					as the source for the blocking analysis of the TornSystem.
				*/

				// Remove the torn variable from the relationship map and take note of the equations
				// which will later require this variable to be substituted (torneqs)
				for(tsi=tsmin.begin(); tsi!=tsmin.end(); ++tsi){
					cerr << "Equation " << *tsi << " (variable " << sys->getVariableName(*tsi) << ")." << endl;
					// remove this tearing variable from the relationship matrix
					for(unsigned r=0; r < n; ++r){
						if(r==*tsi)continue; // don't remove it from the diagonal
						//if(FORTRAN_ARRAY(rel,r,*tsi,n)){
							// cerr << "Removing " << sys->getVariableName(*tsi) << " from eq " << r << endl;
							//FORTRAN_ARRAY(rel,r,*tsi,n) = false;
							set<unsigned> varset=torneqs[*tsi];
							/// add this eqation to torneqs set for this variable
							if(varset.find(r) != varset.end()){
								varset.insert(r);
							}
						//}
					}
				}
				cerr << endl;
	
				cerr << "Creating tornsystem..." << endl;
	
				/// Create a sub-system with these new variables
				tornsystem = new TornSystem();
				for(unsigned i=0; i<n; ++i){
					EqProcess &eq = equations[i];
					tornsystem->inheritVariable(eq.getBaseVariable());
					// the equation we're adding already has the torn equation eliminated
					tornsystem->addExpression(eq.getCurrentEquation());
				}
			
				/*
					The torn system will be solved with a Tarjan Solver to take advantage
					of its now very block-triangular form (hopefully :)
				*/
				tornsolver = new TarjanSolver();
				cerr << "Analysing torn system..." << endl;
				tornsystem->acceptAnalyse(tornsolver);
	
				/*
					The high-level problem (iterating to find the values of the torn variables)
					will be handled by a generic solver, in this case the NNESSolver.
				*/			
				tearingsolver = NNESSolver::Instance();
				
				/**
					The tearing system is the very small problem, just the torn variables.
					Prepare this system now - it will be solved using a standard NNESSolver

					@TODO:
					Or it could even be solved by plugging the residuals into the equations...
					That's an idea for another type of Solver object. Would be good for using on
					the colebrook equation, for example
				*/
				cerr << "Creating TearingSystem..."<< endl;
				tearingsystem = new TearingSystem(*this, *sys, torneqs);
				cerr << "Analyse the TearingSystem (shouldn't involve much)..." << endl;
				tearingsystem->acceptAnalyse(tearingsolver);
				cerr << "TearingSolver::analyseSymbolic: done analysing tearingsystem" << endl;

				
	
				return;
			}else{
				// This block contains no loops so the subsolver will just be a plain NNESSolver
				cerr << "NO LOOPS FOUND" << endl;
			}
		
		}else{
			// If there is only one variable then there can't be any loops, the subsolver will just be a plain NNES again.
			cerr << "THIS BLOCK HAS ONLY ONE VARIABLE" << endl;
		}
	}catch(exception &e){
		stringstream ss;
		ss << "TearingSolver::analyseSymbolic: " << e.what();
		throw runtime_error(ss.str());
	}
}

/**
	The solve process for the TearingSolver is to apply NNES iteration
	to adjust values of the torn variables (by applying NNESSolver to TearingSystem)
	then when the residuals for those torn variables are required, to call the TarjanSolver
	on the TornSystem.
	
	The reason that the torn system is solved using a Tarjan Solver is that
	block decomposition will speed up solution.

	System
	  .
     /|\
      |
      |
    TearingSolver ------>  NNESSolver------> TearingSystem
                                                 |
                                                 v
                                             TarjanSolver --------> TornSystem
*/
void
TearingSolver::solveSymbolic(SymbolicSystem *sys){

	cerr << esc << "[34;1mTearingSolver::solveSymbolic" << esc << "[0m" << endl;
	
	/// create a system
	//TearingSystem *s = new TearingSystem(*this, *sys, torneqs);

	tearingsystem->acceptSolve(tearingsolver);
	
	cerr << "TearingSolver::solve: FOUND SOLUTION" << endl;

	tearingsystem->printsolution();

	//cerr << "Notifying solved variables to base system..."<< endl;
	//sys->copySolution(tornsystem);
	
	sys->printsolution();

}

/**
	For given values of the torn variables, solve the blocks of the system
	and return the error between y and the evaluated torn variables from
	the torn system solution.
*/
void
TearingSolver::getTearResiduals(const double *y, double *res){
	cerr << "TearingSolver::getTearResiduals:..." << endl;

	try{
		
		// substitute the values of the torn variables into our set of equations
		TornEqsMap::const_iterator ti;
		unsigned yi;
		for(ti=torneqs.begin(), yi=0; ti!=torneqs.end(); ++ti, ++yi){
			cerr << "TearingSolver::getTearResiduals: yi = " << yi << endl;
			EqProcess &eq = getEqProcess((*ti).first);
			const SymbolicVariable *var = eq.getBaseVariable();
			cerr << "TearingSolver::getTearResiduals: y[" << yi << "],  " << var->getSymbol() << " = " << y[yi] << endl;
			const set<unsigned> &othereqs = (*ti).second;
			set<unsigned>::const_iterator ei;
			for(ei=othereqs.begin(); ei!=othereqs.end(); ++ei){
				getEqProcess(*ei).subs(var->getSymbol()==y[yi]);
			}
			cerr << "TearingSolver::getTearResiduals: done with yi = " << yi << endl;
		}
	
		// now solve the system which remains
		cerr << "TearingSolver::getTearResiduals: starting up tornsolver..." << endl;
		tornsystem->acceptSolve(tornsolver);
		cerr << "TearingSolver::getTearResiduals: ...tornsolver returned" << endl;
		tornsystem->printsolution();
		cerr << "TearingSolver::getTearResiduals: done printing solution" << endl;
	
		// now pick out the values of the torn variables and return there change as the residual
		for(ti=torneqs.begin(), yi=0; ti!=torneqs.end(); ++ti, ++yi){
			EqProcess &eq = getEqProcess((*ti).first);
			const SymbolicVariable *var = eq.getBaseVariable();
			const GiNaC::symbol &sym = var->getSymbol();
			cerr << "TearingSolver::getTearResiduals: variable " << sym << endl;
			cerr << "TearingSolver::getTearResiduals: y[" << yi << "] = " << y[yi] << endl;
			cerr << "TearingSolver::getTearResiduals: getSolvedVariable(" << yi << ") = " << tornsystem->getSolvedVariable(yi) << endl;
			res[yi] = tornsystem->getSolvedVariable(yi) - y[yi];
		}
	
		tornsystem->printresids(res);
		cerr << "TearingSolver::getTearResiduals:...done" << endl;
	}catch(exception &e){
		stringstream ss;
		ss << "TearingSolver::getTearResiduals: " << e.what() << endl;
		throw runtime_error(ss.str());
	}
}

