/*

nla-ginac - Tarjan-blocking NLA solver with autodiff
Copyright (C) 2005  John Pye

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#ifndef TEARING_H
#define TEARING_H

#include "solver.h"
#include "tarjansystem.h"
#include "eqprocess.h"
#include "tornsystem.h"

using namespace std;

class TearingSystem;

/**
	Tear a system then perform Newton iteration on the
	tear variables.
*/
class TearingSolver : public Solver
{

public:
	TearingSolver();

    virtual const char *getType() const;

	/**
		@TODO: shouldn't a TearingSolver be able to solve non-symbolic systems as well?
		You set the value of the tear variable, then numerically evaluate all the equations
		at those values... nothing there that requires symbolics...
	*/
	virtual void visitSymbolic(SymbolicSystem *s);
	virtual void analyseSymbolic(SymbolicSystem *sys);
	virtual void solveSymbolic(SymbolicSystem *sys);

	virtual void applyLowerBound(const unsigned&,const double&);
	virtual void applyUpperBound(const unsigned&,const double&);

	virtual void getTearResiduals(const double *y, double *res);
	
	virtual EqProcess &getEqProcess(const unsigned &);

private:
	/**
		Equations-in-process: local copies of the equations kept for the purpose
		of eliminating torn variables
	*/
	std::vector<EqProcess> equations;

	/// This will hold the TearingSystem which is prepared during the analysis step
	TearingSystem *tearingsystem;
	
	/// This is the system with the tear made (torn variably values substituted)
	TornSystem *tornsystem;

	/// This is a (standard) solver that will be used to solve the torn equations. A blocking solver will be ideal.
	Solver *tornsolver;

	/// This is a (standard) solver that will hunt for a solution to the torn variables (smalled problem).
	Solver *tearingsolver;

	/**
		torneqs is a map of the form std::map<unsigned,std::set<unsigned> >
		typedef is in symbolicsolver.h
		
		The left side of the mapping is an equation/variable number in the base system
		which corresponds to the torn variable & the equation which is used to estimate
		the residual.

		The right side of the mapping is the equation numbers in the base system into which
		the value of the torn variable must be substituted.
	*/
	TornEqsMap torneqs; ///< equations into which each torn variable must be substituted

};

#endif // TEARING_H
