/*

nla-ginac - Tarjan-blocking NLA solver with autodiff
Copyright (C) 2005  John Pye

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#include <cmath>
#include <stdexcept>
#include <sstream>
#include <iostream>

#include "viscosity.h"
#include "iapws95.h"

/// Dynamic viscosity, mu, [Pa.s]
/**
	Range of validity is entire regions 1,2,3,4

	Reference: J Phys Chem Ref Data, vol 13, no 1, 1984, p. 175 ff.

	@param T temperature / K
	@param rho density / [kg/m3]
	@return Dynamic viscosity [Pa.s]

	@see http://www.iapws.org/relguide/visc.pdf
	@see http://www.iapws.org/relguide/IF97.pdf
*/
Num
Viscosity::dynvisc(const Num &T, const Num &rho){
	Num tau = T / IAPWS::Tstar_visc;
	Num delta = rho / IAPWS::rhostar_visc;
	//std::cerr << "Viscosity::dynvisc: T = " << T << ", rho = " << rho << std::endl;
	return IAPWS::mustar_visc * IAPWS::mu(tau,delta);
}

Num
Viscosity::dynvisc_dT(const Num &T, const Num &rho){
	Num tau = T / IAPWS::Tstar_visc;
	Num delta = rho / IAPWS::rhostar_visc;
	return IAPWS::mustar_visc * IAPWS::mutau(tau,delta) / IAPWS::Tstar_visc;
}

Num
Viscosity::dynvisc_drho(const Num &T, const Num &rho){
	Num tau = T / IAPWS::Tstar_visc;
	Num delta = rho / IAPWS::rhostar_visc;
	return IAPWS::mustar_visc * IAPWS::mudelta(tau,delta) / IAPWS::rhostar_visc;
}

namespace IAPWS
{

	Num mu0(const Num &tau){
		
		Num e = 0;
		for (int i = 0; i < VISC_N0_COUNT; i++) {
			e += VISC_N0[i] / pow(tau, i);
			//printf("NOTE: eta0 = %.5f\n",eta0);
		}
		return (sqrt(tau) / e);
	
	}

	Num mu0tau(const Num &tau){
		Num e = 0;
		for (int i = 0; i < VISC_N0_COUNT; i++) {
			e += VISC_N0[i] / pow(tau, i);
			//printf("NOTE: eta0 = %.5f\n",eta0);
		}
		Num f = 0;
		for(int i=1; i <VISC_N0_COUNT; i++){
			f += -i*VISC_N0[i]*pow(tau,-i-1);
		}
		return 1/2/sqrt(tau)/e - f*sqrt(tau)/pow(e,2);
	}	
	
	/// from the standard
	Num mu1(const Num &tau,const Num&delta){
	
		Num e = 0;
	
		Num tt = 1 / tau - 1;
		Num dd = delta - 1;
		
		for (int i = 0; i < VISC_COUNT; i++) {
			e += VISC_N[i] * pow(tt, VISC_I[i]) * pow(dd, VISC_J[i]);
		}
	
		return (exp(delta * e));
	
	}
	
	/// derived
	Num mu1tau(const Num &tau,const Num&delta){
	
		Num e1 = 0;
	
		Num tt = 1 / tau - 1;
		Num dd = delta - 1;
		
		for (int i = 0; i < VISC_COUNT; i++) {
			e1 += VISC_N[i] * VISC_I[i] * pow(tt, VISC_I[i] - 1) * pow(dd, VISC_J[i]);
		}
		// e1 = diff(e,tt)
		// f = diff(tt,tau)
		Num f = -1/tau/tau;

		Num mu11 = mu1(tau,delta);
	
		return mu11*e1*f;
	}
	
	/// derived
	Num mu1delta(const Num &tau,const Num&delta){
	
		Num e1 = 0;
	
		Num tt = 1 / tau - 1;
		Num dd = delta - 1;
		
		for (int i = 0; i < VISC_COUNT; i++) {
			e1 += VISC_N[i] * pow(tt, VISC_I[i]) * VISC_J[i] * pow(dd, VISC_J[i]-1);
		}
		// e1 = diff(e,tt)
		// f = diff(tt,tau)
		
		Num mu11 = mu1(tau,delta);
	
		return mu11*e1;
	}
	
	
	Num mu2(const Num &tau,const Num&delta){
	
		if (tau >= 0.9970 && delta >= 0.755 && tau < 1.0082 && delta < 1.290) {
			std::stringstream ss;
			ss << "IAPWS::mu2: can't evaluate mu2, too close to critical point, tau = " << tau << ", delta = " << delta << "; this functionality not implemented yet. See IAPWS Release about viscosity.";
			throw std::runtime_error(ss.str());
		} else {
			return 1;
		}
	
	}

	Num mu2tau(const Num &tau,const Num&delta){
		return 0;
	}

	Num mu2delta(const Num &tau,const Num&delta){
		return 0;
	}
	
	/// @see http://www.iapws.org/relguide/IF97.pdf
	Num mu(const Num &tau,const Num&delta){
		//std::cerr << "mu: tau = " << tau << ", delta=" << delta << ", mu0 = " << mu0(tau) << ", mu1 = " << mu1(tau,delta) << ", mu2 = " << mu2(tau,delta);
		double mu = mu0(tau) * mu1(tau,delta) * mu2(tau,delta);
		//std::cerr << " --> " << mu << std::endl;
		return mu;
	}

	Num mutau(const Num &tau,const Num&delta){
		return mu0tau(tau) * mu1(tau,delta) * mu2(tau,delta)
			+ mu0(tau) * mu1tau(tau,delta) * mu2(tau,delta)
           /* + mu0(tau) * mu1(tau,delta) * mu2tau(tau,delta) */;
	}

	Num mudelta(const Num &tau,const Num&delta){
		return 
			+ mu0(tau) * mu1delta(tau,delta) * mu2(tau,delta)
           /* + mu0(tau) * mu1(tau,delta) * mu2delta(tau,delta) */;
	}
}
